"""Tests for transformation processing."""
from typing import Mapping, Tuple, Type, Union, cast

from PIL import Image
import numpy as np
import pandas as pd
from pandas._testing import assert_frame_equal
import pytest
from pytest import fixture

from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.schema import BitfountSchema, TableSchema
from bitfount.data.types import DataSplit
from bitfount.transformations.dataset_operations import (
    NormalizeDataTransformation,
    ScalarAdditionDataTransformation,
    ScalarMultiplicationDataTransformation,
)
from bitfount.transformations.exceptions import (
    InvalidBatchTransformationError,
    MissingColumnReferenceError,
    TransformationApplicationError,
    TransformationProcessorError,
)
from bitfount.transformations.parser import TransformationsParser
from bitfount.transformations.processor import TransformationProcessor
from bitfount.transformations.unary_operations import OneHotEncodingTransformation
from tests.utils import PytestRequest
from tests.utils.helper import create_dataset, integration_test, unit_test


def parse_and_process_data(
    data: pd.DataFrame, schema: TableSchema, transformations_yaml: str
) -> Tuple[pd.DataFrame, TableSchema]:
    """Parses transformations YAML and applies them to data."""
    parser = TransformationsParser()
    transformations, cols = parser.parse(transformations_yaml)
    processor = TransformationProcessor(transformations, schema, cols)
    data = processor.transform(data)
    return data, cast(TableSchema, processor.schema)


IntegerTypes = Tuple[
    Type[int], Type[np.int32], Type[np.int64], pd.Int32Dtype, pd.Int64Dtype
]


class TestTransformationProcessor:
    """Tests for TransformationProcessor class."""

    @fixture(scope="function")
    def data_and_schema(
        self, request: PytestRequest
    ) -> Tuple[pd.DataFrame, TableSchema]:
        """Dataset and schema to be used for tests."""
        data = create_dataset()
        datasource = DataFrameSource(data)
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name="test")
        datasource.load_data()
        return datasource.data, schema.tables[0]

    @fixture
    def integer_types(self) -> IntegerTypes:
        """Fixture determining what counts as an integer type."""
        return (int, np.int32, np.int64, pd.Int32Dtype(), pd.Int64Dtype())

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_addition_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests addition transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Add:
            name: A_and_B_sum
            output: true
            arg1: col:A
            arg2: c:B
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_B_sum" in data.columns
        assert data["A_and_B_sum"].sum() == data.A.sum() + data.B.sum()

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_comparison_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests comparison transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Compare:
            name: A_and_B_comp
            output: true
            arg1: col:A
            arg2: c:B
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_B_comp" in data.columns
        comp = sum((data.A > data.B).to_list()) - sum((data.B > data.A).to_list())
        assert data["A_and_B_comp"].sum() == comp

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_comparison_transformation_with_constant(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests comparison transformation works correctly with a constant."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Compare:
            name: A_and_2_comp
            output: true
            arg1: col:A
            arg2: 2
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_2_comp" in data.columns
        comp = sum((data.A > 2).to_list()) - sum((data.A < 2).to_list())
        assert data["A_and_2_comp"].sum() == comp

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_division_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests division transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Divide:
            name: A_and_B_div
            output: true
            arg1: col:A
            arg2: c:B
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_B_div" in data.columns
        assert int(data["A_and_B_div"].sum()) == int(
            np.sum((data.A / data.B).to_numpy())
        )

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_inclusion_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests inclusion transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - In:
            name: a_in_I
            output: true
            in_str: a
            arg: c:I
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "a_in_I" in data.columns
        assert sum(data["a_in_I"].to_list()) == sum((data.I == "a").to_list())

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_multiplication_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests multiplication transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Multiply:
            name: A_and_B_prod
            output: true
            arg1: col:A
            arg2: c:B
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_B_prod" in data.columns
        assert data["A_and_B_prod"].sum() == data.A @ data.B

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_one_hot_encoding_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests one-hot encoding transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - OneHotEncode:
            name: ohe_I
            output: true
            arg: col:I
            values: [e,f,g]
        """
        original_data_len = len(data.index)
        data, _ = parse_and_process_data(data, schema, test_yaml)

        # Check all one hot columns are included
        one_hot_cols = [f"ohe_I_{letter}" for letter in [*"efg", "UNKNOWN"]]
        for col in one_hot_cols:
            assert col in data.columns

        # Check no others are
        for col in data.columns:
            if col.startswith("ohe"):
                assert col in one_hot_cols

        # Check additional columns have been correctly concatenated
        assert len(data.index) == original_data_len

    @fixture
    def ohe_mini_dataset_and_results(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Provides a small dataset, results and dataframe type for tests.

        Generates a (4,3) test dataframe for testing one-hot encoding and a
        corresponding results dataframe (4,7) which encodes values [1, 2] from
        "col1".

        Also returns the type of the dataframes.
        """
        data_in = {
            "col1": [1, 2, 3, None],
            "col2": [4, 5, 6, None],
            "col3": [7, 8, 9, None],
        }
        results = {
            "col1": [1, 2, 3, None],
            "col2": [4, 5, 6, None],
            "col3": [7, 8, 9, None],
            "col1_1": [1, 0, 0, 0],
            "col1_2": [0, 1, 0, 0],
            "col1_UNKNOWN": [0, 0, 1, 0],
        }

        return (
            pd.DataFrame(data=data_in),
            pd.DataFrame(data=results),
        )

    @unit_test
    def test__do_one_hot_encoding(
        self,
        ohe_mini_dataset_and_results: Tuple[pd.DataFrame, pd.DataFrame],
    ) -> None:
        """Tests the one hot encoding method directly.

        Checks that values are being encoded correctly.
        """
        data_in, expected_results = ohe_mini_dataset_and_results

        ohe_transform = OneHotEncodingTransformation(arg="c:col1", raw_values=[1, 2])
        results = TransformationProcessor._do_one_hot_encoding(data_in, ohe_transform)

        assert_frame_equal(results, expected_results, check_dtype=False)

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_subtraction_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests subtraction transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - Subtract:
            name: A_and_B_sub
            output: true
            arg1: col:A
            arg2: c:B
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "A_and_B_sub" in data.columns
        assert data["A_and_B_sub"].sum() == data.A.sum() - data.B.sum()

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_cleandata_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests clean data transformation works correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - CleanData:
            name: cleandata
            cols: col:Date
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "cleandata" not in data.columns
        assert data["Date"].isnull().sum() == 0

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_cleandata_transformation_no_args(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests clean data transformation with no arguments (i.e. all cols)."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - CleanData
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert data["Date"].isnull().sum() == 0

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_normalize_transformation(
        self,
        data_and_schema: Tuple[pd.DataFrame, TableSchema],
        integer_types: IntegerTypes,
    ) -> None:
        """Tests normalize transformation works."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - NormalizeData:
            name: blah
            cols:
                - col:A
                - col:B
        """
        # Assert columns
        for col in ["A", "B", "C", "D"]:
            assert schema.features["continuous"][col].dtype in integer_types
        data, schema = parse_and_process_data(data, schema, test_yaml)
        assert round(data.A.mean()) == 0
        assert round(data.A.std()) == 1
        assert round(data.B.mean()) == 0
        assert round(data.B.std()) == 1

        # Assert that only columns A and B have been normalized
        assert schema.features["continuous"]["A"].dtype == np.float64
        assert schema.features["continuous"]["B"].dtype == np.float64
        assert schema.features["continuous"]["C"].dtype in integer_types
        assert schema.features["continuous"]["D"].dtype in integer_types

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_normalize_transformation_on_categorical_column(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests exception raised if normalize called on categorical column."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - NormalizeData:
            name: blah
            cols:
                - col:I
        """
        with pytest.raises(TransformationApplicationError):
            parse_and_process_data(data, schema, test_yaml)

    @unit_test
    def test_normalize_transformation_with_list_of_cols(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests normalization works when instantiated with a list of columns."""
        data, schema = data_and_schema
        transform = NormalizeDataTransformation(cols=["A", "B"])
        processor = TransformationProcessor([transform], schema)
        data = processor.transform(data)
        assert round(data["A"].mean()) == 0
        assert round(data["A"].std()) == 1
        assert round(data["B"].mean()) == 0
        assert round(data["B"].std()) == 1

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_normalize_transformation_no_args(
        self,
        data_and_schema: Tuple[pd.DataFrame, TableSchema],
        integer_types: IntegerTypes,
    ) -> None:
        """Tests NormalizeData transformation when no args supplied (all cols)."""
        data, schema = data_and_schema
        data["E"] = 2 * data["E"]
        data["F"] = 5 * data["F"]
        data["G"] = 10 * data["G"]
        data["H"] = 3 * data["H"]
        # Check that the mean is no longer 0
        for feature in ["E", "F", "G", "H"]:
            assert round(data[feature].mean()) != 0
        test_yaml = """
        Transformations:
        - NormalizeData
        """
        data, schema = parse_and_process_data(data, schema, test_yaml)
        # Check that int columns remain unchanged
        for feature in ["A", "B", "C", "D"]:
            assert schema.features["continuous"][feature].dtype in integer_types
        # Check that the float columns have been normalized
        for feature in ["E", "F", "G", "H"]:
            assert round(data[feature].mean()) == 0
            assert round(data[feature].std()) == 1

    @unit_test
    def test_errors_if_transformation_same_name_as_col(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests an error is raised when a transformation name matches a column."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - add:
            name: A  # this matches a column name
            arg1: col:C
            arg2: col:D
        """
        with pytest.raises(TransformationApplicationError) as tpe:
            parse_and_process_data(data, schema, test_yaml)
        assert tpe.value.errors == [
            "Output column A, from transformation A, "
            "clashes with an existing data column name."
        ]

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_errors_if_multi_column_transformation_out_col_same_as_col(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests an error is raised when an MCO output column name matches a column."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - onehotencode:
            name: A  # this matches a column name, but doesn't matter
            arg: col:B
            values: {1: "C", 2: "D"}
        """
        # Rename data column to be problematic
        data = data.rename(columns={"C": "A_C", "D": "A_D"})

        with pytest.raises(TransformationApplicationError) as tpe:
            parse_and_process_data(data, schema, test_yaml)

        assert tpe.value.errors == [
            "Output column A_C, from transformation A, "
            "clashes with an existing data column name.",
            "Output column A_D, from transformation A, "
            "clashes with an existing data column name.",
        ]

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_multi_transformation(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the application of multiple transformations."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - divide:
            name: A_and_B_div
            output: false
            arg1: col:A
            arg2: c:B
        - multiply:
            name: percentage
            output: true
            arg1: tran:A_and_B_div
            arg2: 100
        - divide:
            name: reciprocal
            output: true
            arg2: 1
            arg1: tran:A_and_B_div
        - in:
            in_str: "a"
            arg: column:I
            output: true
        """
        original_col_length = len(data.columns)
        data, _ = parse_and_process_data(data, schema, test_yaml)
        assert "reciprocal" in data.columns
        assert "percentage" in data.columns
        assert "A_and_B_div" not in data.columns
        assert original_col_length + 3 == len(data.columns)
        assert data.iloc[:, -1:].columns[0][:3] == "in_"

    @pytest.mark.parametrize(
        "data_and_schema",
        [
            pytest.param(True, marks=integration_test),
            pytest.param(False, marks=unit_test),
        ],
        indirect=["data_and_schema"],
    )
    def test_multi_column_output_drop_cols(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests that multicolumn outputs are all dropped correctly."""
        data, schema = data_and_schema
        test_yaml = """
        Transformations:
        - OneHotEncode:
            name: ohe_I
            output: false
            arg: col:I
            values: [e,f,g]
        """
        data, _ = parse_and_process_data(data, schema, test_yaml)
        # Check no column with ohe_I in it exists
        for col in data.columns:
            assert not col.startswith("ohe_I")

    @unit_test
    def test_schema_invalid_columns(self) -> None:
        """Tests exception raised if referenced column doesn't exist."""
        test_yaml = """
        Transformations:
        - divide:
            name: A_and_B_div
            output: true
            arg1: col:col_that_doesntexist
            arg2: c:second_nonexistent_col
        """
        data = create_dataset()
        schema = BitfountSchema(DataFrameSource(data), table_name="test")
        parser = TransformationsParser()
        transformations, cols = parser.parse(test_yaml)
        processor = TransformationProcessor(transformations, schema.tables[0], cols)
        with pytest.raises(MissingColumnReferenceError) as tpe:
            processor.transform(data)
        assert tpe.value.errors == [
            f"Reference to non-existent column: {c}"
            for c in ("col_that_doesntexist", "second_nonexistent_col")
        ]

    @unit_test
    def test_schema_invalid_operation(self) -> None:
        """Tests exception raised if transformation operation invalid."""
        test_yaml = """
        Transformations:
        - divide:
            name: A_and_B_div
            output: false
            arg1: col:A
            arg2: c:I
        """
        data = create_dataset()
        schema = BitfountSchema(DataFrameSource(data), table_name="test")
        parser = TransformationsParser()
        transformations, cols = parser.parse(test_yaml)
        processor = TransformationProcessor(transformations, schema.tables[0], cols)
        original_col_length = len(data.columns)
        with pytest.raises(TransformationApplicationError):
            processor.transform(data)
        assert "A_and_B_div" not in data.columns
        assert original_col_length == len(data.columns)

    @unit_test
    def test_image_transformation(self) -> None:
        """Tests image transformation works correctly."""
        test_yaml = """
        Transformations:
        - albumentations:
            arg: col:col1
            output: True
            step: train
            transformations:
                - HorizontalFlip:
                    p: 0.5
                - RandomBrightnessContrast
        """
        parser = TransformationsParser()
        tfms, _ = parser.parse(test_yaml)
        proc = TransformationProcessor(tfms)
        image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
        transformed_image = proc.batch_transform(image, step=DataSplit.TRAIN)
        assert isinstance(transformed_image, np.ndarray)
        assert transformed_image.shape == image.shape

    @unit_test
    def test_image_transformation_step_is_respected(self) -> None:
        """Tests that transformation is not applied if step does not match."""
        test_yaml = """
        Transformations:
        - albumentations:
            arg: col:col1
            output: True
            step: train
            transformations:
                - HorizontalFlip:
                    p: 0.5
                - RandomBrightnessContrast
        """
        parser = TransformationsParser()
        tfms, _ = parser.parse(test_yaml)
        proc = TransformationProcessor(tfms)
        image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
        transformed_image = proc.batch_transform(image, step=DataSplit.VALIDATION)

        # Assert that the 'transformed' image is exactly the same
        np.testing.assert_array_equal(transformed_image, image)

    @unit_test
    def test_batch_transform_raises_value_error_with_non_batch_transformation(
        self,
    ) -> None:
        """Tests that `batch_tranform` can only be used with batch transformations."""
        test_yaml = """
        Transformations:
        - divide:
            name: A_and_B_div
            output: false
            arg1: col:A
            arg2: c:I
        """
        parser = TransformationsParser()
        tfms, _ = parser.parse(test_yaml)
        proc = TransformationProcessor(tfms)
        image = np.array(Image.new("RGB", size=(50, 50), color=(100, 0, 255)))
        with pytest.raises(InvalidBatchTransformationError):
            proc.batch_transform(image, step=DataSplit.TRAIN)

    @unit_test
    @pytest.mark.parametrize(
        "scalar",
        [pytest.param(2), pytest.param(0.1), pytest.param({"A": 0.2, "B": 10, "C": 4})],
    )
    def test_scalar_multiplication(
        self,
        data_and_schema: Tuple[pd.DataFrame, TableSchema],
        scalar: Union[int, float, Mapping[str, Union[int, float]]],
    ) -> None:
        """Tests the scalar multiplication is applied correctly."""
        data, schema = data_and_schema
        orig_data = data.copy()
        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        data = processor.transform(data)
        if isinstance(scalar, (int, float)):
            for feature in ["E", "F", "G", "H"]:
                assert (orig_data[feature] * scalar).to_list() == data[
                    feature
                ].to_list()
        else:
            for feature in ["A", "B", "C"]:
                assert (orig_data[feature] * scalar[feature]).to_list() == data[
                    feature
                ].to_list()

    @unit_test
    def test_scalar_multiplication_raises_error_non_existing_column(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar multiplication fails if col doesn't exist."""
        scalar = {"X": 1.5, "Z": 0.2}
        data, schema = data_and_schema
        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        with pytest.raises(MissingColumnReferenceError) as tpe:
            processor.transform(data)
        assert (
            tpe.value.errors.sort()
            == [f"Reference to non-existent column: {c}" for c in scalar.keys()].sort()
        )

    @unit_test
    def test_scalar_multiplication_raises_non_numeric_column(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar multiplication fails non-numeric column."""
        scalar = {"Date": 1.5}
        data, schema = data_and_schema
        scalar_transform = ScalarMultiplicationDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        with pytest.raises(TransformationApplicationError):
            processor.transform(data)

    @unit_test
    def test_scalar_multiplication_raises_error_wrong_scalar(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar multiplication if scalar is an incorrect format."""
        scalar = [1, 2, 3]
        with pytest.raises(TransformationApplicationError):
            ScalarMultiplicationDataTransformation(scalar=scalar)  # type: ignore[arg-type] # Reason: incorrect assignment to check if error is raised. # noqa: B950

    @unit_test
    @pytest.mark.parametrize(
        "scalar",
        [pytest.param(2), pytest.param(0.1), pytest.param({"A": 0.2, "B": 10, "C": 4})],
    )
    def test_scalar_addition(
        self,
        data_and_schema: Tuple[pd.DataFrame, TableSchema],
        scalar: Union[int, float, Mapping[str, Union[int, float]]],
    ) -> None:
        """Tests the scalar addition is applied correctly."""
        data, schema = data_and_schema
        orig_data = data.copy()
        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        data = processor.transform(data)
        if isinstance(scalar, (int, float)):
            for feature in ["E", "F", "G", "H"]:
                assert (orig_data[feature] + scalar).to_list() == data[
                    feature
                ].to_list()
        else:
            for feature in ["A", "B", "C"]:
                assert (orig_data[feature] + scalar[feature]).to_list() == data[
                    feature
                ].to_list()

    @unit_test
    def test_scalar_addition_raises_error_non_existing_column(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar addition fails if col doesn't exist."""
        scalar = {"X": 1.5, "Z": 0.2}
        data, schema = data_and_schema
        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        with pytest.raises(MissingColumnReferenceError) as tpe:
            processor.transform(data)
        assert (
            tpe.value.errors.sort()
            == [f"Reference to non-existent column: {c}" for c in scalar.keys()].sort()
        )

    @unit_test
    def test_scalar_addition_raises_error_wrong_scalar(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar addition if scalar is an incorrect format."""
        scalar = [1, 2, 3]
        with pytest.raises(TransformationApplicationError):
            ScalarAdditionDataTransformation(
                scalar=scalar  # type: ignore[arg-type] # Reason: incorrect assignment to check if error is raised. # noqa: B950
            )

    @unit_test
    def test_scalar_addition_raises_non_numeric_column(
        self, data_and_schema: Tuple[pd.DataFrame, TableSchema]
    ) -> None:
        """Tests the scalar addition fails non-numeric column."""
        scalar = {"Date": 1.5}
        data, schema = data_and_schema
        scalar_transform = ScalarAdditionDataTransformation(scalar=scalar)
        processor = TransformationProcessor([scalar_transform], schema)
        with pytest.raises(TransformationApplicationError):
            processor.transform(data)


@unit_test
class TestTransformationProcessorError:
    """Tests for TransformationProcessorError exception."""

    def test_single_error_arg(self) -> None:
        """Tests creation of TransformationProcessorError from single error."""
        with pytest.raises(TransformationProcessorError) as tpe:
            raise TransformationProcessorError("test")
        assert tpe.value.errors == ["test"]

    def test_multiple_error_arg(self) -> None:
        """Tests creation of TransformationProcessorError with multiple errors."""
        with pytest.raises(TransformationProcessorError) as tpe:
            raise TransformationProcessorError(["test", "test2"])
        assert tpe.value.errors == ["test", "test2"]

    def test_str(self) -> None:
        """Tests string representation of TransformationProcessorError."""
        with pytest.raises(TransformationProcessorError) as tpe:
            raise TransformationProcessorError(["test", "test2"])
        assert str(tpe.value) == "Errors: \ntest\ntest2"



"""Test config settings for the mocked end-to-end tests."""
from typing import Callable, Dict, Tuple

import flask
from flask import Flask, jsonify
from flask import request as flask_request
from pytest import fixture

from bitfount.hub.types import (
    _HubFailureResponseJSON,
    _HubSuccessResponseJSON,
    _PodDetailsResponseJSON,
    _PodRegistrationFailureJSON,
    _PodRegistrationResponseJSON,
)
from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL
from tests.utils import PytestRequest


@fixture
def s3_upload_url_builder(
    s3_netloc_subdomain: str,
) -> Callable[[str], _S3PresignedPOSTURL]:
    """Builds an S3 upload url with the specified pod name at the end."""

    def _create(pod_name: str) -> _S3PresignedPOSTURL:
        return _S3PresignedPOSTURL(
            f"https://an-s3-upload-url{s3_netloc_subdomain}.s3.com/some-resource/{pod_name}"
        )

    return _create


@fixture
def s3_download_url_builder(
    s3_netloc_subdomain: str,
) -> Callable[[str], _S3PresignedURL]:
    """Builds an S3 download url with the specified pod name at the end."""

    def _create(pod_name: str) -> _S3PresignedURL:
        return _S3PresignedURL(
            f"https://an-s3-download-url{s3_netloc_subdomain}.s3.com/some-resource/{pod_name}"
        )

    return _create


@fixture
def app(
    modeller_public_key_string: str,
    request: PytestRequest,
    s3_download_url_builder: Callable[[str], _S3PresignedURL],
    s3_upload_fields: _S3PresignedPOSTFields,
    s3_upload_url_builder: Callable[[str], _S3PresignedPOSTURL],
) -> Flask:
    """Creates the Bitfount Hub API."""
    register_pod_fail, access_request_fail, get_pod_fail = (
        False,
        False,
        False,
    )
    if hasattr(request, "param"):
        (
            register_pod_fail,
            access_request_fail,
            get_pod_fail,
        ) = request.param

    api = Flask(__name__)

    pod_public_keys: Dict[str, str] = {}

    @api.route("/")
    def index() -> str:
        """Top level of the app. Just used to retrieve the URL."""
        return ""  # needs to return _something_

    @api.route("/api/pods", methods=["POST", "PATCH"])
    def register_pod() -> Tuple[flask.Response, int]:
        if flask_request.method == "POST":
            if register_pod_fail:
                # 400 POST response has { success, errorMessage, alreadyExisted } JSON
                post_failure_response: _PodRegistrationFailureJSON = {
                    "success": False,
                    "errorMessage": "Fail!",
                    "alreadyExisted": False,
                }
                return (
                    jsonify(post_failure_response),
                    400,
                )
            content = flask_request.get_json()
            pod_name: str = content.get("name")
            pod_public_key: str = content.get("podPublicKey")
            pod_public_keys[pod_name] = pod_public_key
            upload_url = s3_upload_url_builder(pod_name)
            # 200/201 POST response has { success, alreadyExisted, message,
            # uploadUrl, uploadFields } JSON
            post_success_response: _PodRegistrationResponseJSON = {
                "success": True,
                "message": "Success!",
                "alreadyExisted": False,
                "uploadUrl": upload_url,
                "uploadFields": s3_upload_fields,
            }
            return jsonify(post_success_response), 200
        elif flask_request.method == "PATCH":
            content = flask_request.get_json()
            pod_name = content.get("name")
            pod_public_key = content.get("podPublicKey")
            pod_public_keys[pod_name] = pod_public_key
            # 200 PATCH response has { success, message } JSON
            success_response: _HubSuccessResponseJSON = {
                "success": True,
                "message": "Success!",
            }
            return (
                jsonify(success_response),
                200,
            )
        raise TypeError(f"Unrecognised HTTP method {flask_request.method}")

    @api.route("/api/pods/<username>/<pod_name>", methods=["GET"])
    def get_pod(username: str, pod_name: str) -> Tuple[flask.Response, int]:
        if get_pod_fail:
            failure_response: _HubFailureResponseJSON = {
                "success": False,
                "errorMessage": "Fail!",
            }
            return jsonify(failure_response), 500

        pod_details_json: _PodDetailsResponseJSON = {
            "podIdentifier": f"{username}/{pod_name}",
            "podName": pod_name,
            "podDisplayName": pod_name,
            "podPublicKey": pod_public_keys[pod_name],
            "accessManagerPublicKey": "not_a_real_am_key",
            "description": pod_name,
            "schemaStorageKey": "schema_storage_key",
            "isOnline": True,
            "providerUserName": "providerUserName",
            "visibility": "public",
            "schemaDownloadUrl": s3_download_url_builder(pod_name),
        }
        return jsonify(pod_details_json), 200

    api.testing = True

    return api



"""Fixtures for components that are from the transport layer."""
import asyncio
from asyncio import Queue, QueueEmpty
import threading
from typing import AsyncIterator, Callable, Iterable, Optional, Protocol, Type, Union
from unittest.mock import AsyncMock, MagicMock, Mock, create_autospec

from grpc import RpcError, StatusCode
from pytest import fixture
from pytest_mock import MockerFixture
from typing_extensions import TypeAlias

from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _MessageService,
)
from tests.utils import PytestRequest
from tests.utils.helper import end_to_end_mocks_test, integration_test, unit_test

MessageOrException: TypeAlias = Union[
    _BitfountMessage, BaseException, Type[BaseException]
]


@fixture
def mock_grpc_insecure_channel(mocker: MockerFixture) -> Mock:
    """Mock out grpc.insecure_channel() import in the config module.

    As aio is imported by `from grpc import aio` in the config module we need to
    mock it out that way.
    """
    mock_insecure_channel: Mock = mocker.patch(
        "grpc.aio.insecure_channel", autospec=True
    )
    return mock_insecure_channel


@fixture
def mock_message_service_stub(mocker: MockerFixture) -> Mock:
    """Mock out MessageServiceStub import in the config module."""
    mock_message_service_stub: Mock = mocker.patch(
        "bitfount.federated.transport.config.MessageServiceStub", autospec=True
    )
    return mock_message_service_stub


@fixture
def mock_message_service() -> Mock:
    """Returns a mocked message service."""
    mock_message_service: Mock = create_autospec(_MessageService, instance=True)
    return mock_message_service


class PollForMessagesPatcher(Protocol):
    """Callback protocol for the poll_for_messages patcher."""

    def __call__(
        self,
        messages: Iterable[MessageOrException],
        finish: bool = False,
    ) -> MagicMock:
        """Call signature."""
        ...


@fixture
def patch_poll_for_messages(
    # this should be another fixture in the scope of the tests being run
    mock_message_service: Mock,
) -> PollForMessagesPatcher:
    """Returns a function that can be used to assign messages to yield."""

    def poll_for_messages_patcher(
        messages: Iterable[MessageOrException],
        finish: bool = False,
    ) -> MagicMock:
        """Assigns an iterable of messages to the yield of poll_for_messages.

        If finish is True the poll_for_messages will end after the messages in
        the iterable, otherwise it will "wait" indefinitely as though waiting for
        another message from the message service.
        """
        # Assign these to poll_for_messages which is the underlying method we
        # expect to yield messages.

        async def _mock_poll_for_messages(
            mailbox_id: str,
            stop_event: threading.Event,
        ) -> AsyncIterator[_BitfountMessage]:
            queue_: Queue[MessageOrException] = Queue()
            for message in messages:
                queue_.put_nowait(message)

            while not stop_event.is_set():
                try:
                    item = queue_.get_nowait()

                    # If an exception, raise it
                    if (
                        isinstance(item, BaseException)
                        or isinstance(item, type)
                        and issubclass(item, BaseException)
                    ):
                        raise item

                    # Otherwise yield it
                    yield item
                except QueueEmpty:
                    if finish:
                        return
                    else:
                        await asyncio.sleep(0)

        # Deliberately use MagicMock rather than AsyncMock here because we want
        # to force the wrapped method to return rather than returning a coroutine
        # wrapping the coroutine.
        mock_poll_for_messages = MagicMock(wraps=_mock_poll_for_messages)
        mock_message_service.poll_for_messages = mock_poll_for_messages
        return mock_poll_for_messages

    return poll_for_messages_patcher


class RpcErrorMaker(Protocol):
    """Callback Protocol defining a factory method for RPC errors."""

    def __call__(self, msg: str = ..., code: StatusCode = ...) -> RpcError:
        """Call the factory method."""
        ...


@fixture
def rpc_error_maker() -> RpcErrorMaker:
    """Creates factory method for creating gRPC error objects."""

    def _maker(
        msg: str = "Unknown error", code: StatusCode = StatusCode.UNKNOWN
    ) -> RpcError:
        """Create gRPC error instance.

        Args:
            msg: The error message to be associated with the error.
            code: The status code of the error.

        Returns:
            RpcError instance.
        """
        # Create fake RpcError, manually set code/details methods as valid RpcError
        # can't be created directly in Python code.
        err = RpcError(msg)
        err.code = lambda: code  # type: ignore[assignment] # Reason: see comment # noqa: B950
        err.details = lambda: msg  # type: ignore[assignment] # Reason: see comment # noqa: B950
        return err

    return _maker


@fixture
def rpc_error(rpc_error_maker: RpcErrorMaker) -> RpcError:
    """Creates an RpcError exception to use in mocking error message retrieval.

    Error code is "UNKNOWN".
    """
    return rpc_error_maker()


@fixture
def mock_message_timestamps(mocker: MockerFixture) -> Callable[[Iterable[str]], Mock]:
    """Returns a callable that will mock out message_service._current_time.

    Each call to _current_time() will instead return the next string in the supplied
    iterable.
    """

    def _message_timestamp_mocker(fake_timestamps: Iterable[str]) -> Mock:
        # Because of how dataclass default_factory works, patching
        # message_service._current_time won't be sufficient as it's already bound
        # to the dataclass. Instead we have to patch out the inner calls to datetime.
        mock_datetime: Mock = mocker.patch(
            "bitfount.federated.transport.message_service.datetime", autospec=True
        )
        mock_datetime.now.return_value.isoformat.side_effect = fake_timestamps
        return mock_datetime

    return _message_timestamp_mocker


@fixture(autouse=True)
def remove_grpc_retry_backoff_sleep(
    mocker: MockerFixture, request: PytestRequest
) -> Optional[AsyncMock]:
    """Removes backoff sleep for the gRPC auto_retry functionality.

    As we don't want to slow down tests unnecessarily, this fixture mocks out
    the sleep that would normally occur in the automatic retry of web interactions.
    """
    # Only mock it out for tests that don't actually perform web interactions
    # i.e. not e2e or tutorial tests
    if any(
        mark.name in request.keywords
        for mark in (unit_test, integration_test, end_to_end_mocks_test)
    ):
        # Need to patch entire imported module but only want to mock out `.sleep`
        # calls so set that explicitly to a Mock that isn't wrapping
        mock_async: Mock = mocker.patch(
            "bitfount.federated.transport.utils.asyncio", wraps=asyncio
        )
        mock_async.sleep = AsyncMock()
        return mock_async.sleep
    else:
        return None



"""Useful objects/functions for federated learning.

This is primarily intended for use by modules outside of the `federated` package.
It cannot be imported by most modules in the `federated` package because it would
introduce circular imports.
"""
from __future__ import annotations

import importlib
import inspect
import pkgutil
import sys
from types import ModuleType
from typing import Any, Dict, List, Mapping, Type, Union, cast

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.hashes import HashAlgorithm
import pandas as pd

import bitfount.federated.aggregators.base as aggregators
import bitfount.federated.algorithms.base as algorithms
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.mixins import _DistributedModelMixIn
import bitfount.federated.protocols.base as protocols
import bitfount.models.base_models as model_structures
from bitfount.models.base_models import (
    NeuralNetworkModelStructure,
    NeuralNetworkPredefinedModel,
    _BaseModel,
)
import bitfount.models.models as core_models
from bitfount.types import DistributedModelProtocol

logger = _get_federated_logger(__name__)

# This is a read-only dictionary mapping the name of an aggregator to the class itself
_AGGREGATORS: Mapping[
    str, Type[aggregators._BaseAggregatorFactory]
] = aggregators.registry


# This is a read-only dictionary mapping the name of an algorithm to the class itself
_ALGORITHMS: Mapping[str, Type[algorithms._BaseAlgorithmFactory]] = algorithms.registry


# This is a read-only dictionary mapping the name of a protocol to the class itself
_PROTOCOLS: Mapping[str, Type[protocols._BaseProtocolFactory]] = protocols.registry


_MODEL_STRUCTURES: Dict[
    str, Type[Union[NeuralNetworkPredefinedModel, NeuralNetworkModelStructure]]
] = {
    name: class_
    for name, class_ in vars(model_structures).items()
    if inspect.isclass(class_)
    and issubclass(class_, (NeuralNetworkPredefinedModel, NeuralNetworkModelStructure))
    and not inspect.isabstract(class_)
}


def _load_models_from_module(module: ModuleType) -> Dict[str, Type[_BaseModel]]:
    """Load all concrete classes subclassing _BaseModel from a module.

    Args:
        module (ModuleType): The module to load models from.

    Returns:
        Dict[str, Type[_BaseModel]]: A dict of class name to class for all models found
    """
    found_models: Dict[str, Type[_BaseModel]] = {}

    # Load any concrete classes that extend DistributedModelMixIn and _BaseModel
    for cls_name, class_ in vars(module).items():
        if (
            inspect.isclass(class_)
            and issubclass(class_, _BaseModel)
            and not inspect.isabstract(class_)
        ):
            found_models[cls_name] = class_

    return found_models


def _load_backend_models() -> Dict[str, Type[_BaseModel]]:
    """Load model definitions from backends.

    Load all models defined in model.py modules for backends in the
    bitfount.backends package.

    Returns:
        Dict[str, Type[_BaseModel]]: A dict of class name to class for all models found
    """
    found_models: Dict[str, Type[_BaseModel]] = {}

    try:
        import bitfount.backends
    except ModuleNotFoundError:
        logger.warning("Unable to import bitfount.backends; does it exist?")
        return {}

    backend_path = bitfount.backends.__path__
    backend_prefix = f"{bitfount.backends.__name__}."

    # Find all top-level modules in the backend directory
    for _finder, name, _ispkg in pkgutil.iter_modules(backend_path, backend_prefix):
        models_module = None
        try:
            # Import their models.py file
            models_module = f"{name}.models.models"
            module = importlib.import_module(models_module)
            found_models.update(_load_models_from_module(module))
        except ImportError as ie:
            loc = models_module if models_module else name
            logger.error(f"Unable to import models from backend {loc}: {ie}")
    return found_models


def hash(record: Any, hash_function: HashAlgorithm) -> int:
    """Compute the hash of an individual record.

    Args:
        record: The record to hash.
        hash_function: the hash function to use.

    Returns:
        An integer corresponding to the hashed value of the record.
    """
    # Encode the string to bytes
    encoded_x = str(record).encode()
    # Compute the hash
    digest = hashes.Hash(hash_function)
    digest.update(encoded_x)
    hashed_value = digest.finalize()
    # Return the hash as an integer
    return int.from_bytes(hashed_value, sys.byteorder)


def hash_set(
    dataset: Union[pd.DataFrame, List[Any], pd.Series],
    hash_function: HashAlgorithm,
) -> List[int]:
    """Hash the dataset with the given hash function.

    Returns:
        The hashed dataset as a list of integers.
    """
    if isinstance(dataset, pd.DataFrame):
        hashed_list = []
        for _, row in dataset.iterrows():
            # Remove the pd.dataframe index column from each row
            # Each row is a series now, if there is a column 'name'
            # in the series it should not be deleted.
            row.name = None
            hashed_list.append(hash(row, hash_function))
        return hashed_list
    elif isinstance(dataset, (list, pd.Series)):
        return [hash(item, hash_function) for item in dataset]


def _keys_exist(element: Dict, *keys: str) -> bool:
    """Check if *keys (nested) exists in `element` (dict).

    Args:
        element: The dictionary on which the search is done.
        *keys: Strings that represent the keys to the nested dictionary.
    """
    if not isinstance(element, dict):
        raise AttributeError("keys_exists() expects dict as first argument.")
    if len(keys) == 0:
        raise AttributeError("keys_exists() expects at least two arguments, one given.")

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


# This is a dictionary mapping the name of a model to the model class itself
# Load all models in main models.py
_MODELS: Dict[str, Type[_BaseModel]] = _load_models_from_module(core_models)
# Load any models in a backend's models.py
_MODELS.update(_load_backend_models())

# This is a dictionary mapping the name of a distributed model to the model class itself
_DISTRIBUTED_MODELS: Dict[str, Type[DistributedModelProtocol]] = {}

# We take a subset of MODELS that additionally subclass DistributedModelMixIn
for model_name, model_class in _MODELS.items():
    if issubclass(model_class, _DistributedModelMixIn):
        _DISTRIBUTED_MODELS[model_name] = cast(
            Type[DistributedModelProtocol], model_class
        )



"""Tests for dataset transformations."""
from typing import Type

from marshmallow import ValidationError
import pytest

from bitfount.transformations.base_transformation import Transformation
from bitfount.transformations.dataset_operations import (
    CleanDataTransformation,
    DatasetTransformation,
    NormalizeDataTransformation,
)
from tests.bitfount.transformations.transformation_test_helpers import (
    gen_name_fails,
    gen_name_test,
    registration_test,
)
from tests.utils.helper import unit_test


@unit_test
class TestDatasetTransformation:
    """Tests for the DatasetTransformation abstract transformation."""

    def test__gen_name_fails(self) -> None:
        """Tests that name generation fails."""
        gen_name_fails(DatasetTransformation)

    def test_schema_loads(self) -> None:
        """Tests schema load works correctly."""
        s = DatasetTransformation.schema()
        t = s.loads(
            """
            name: test
            output: True
            cols: col:A
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.cols == "col:A"

    def test_schema_loads_default_args(self) -> None:
        """Tests the state of the default args on schema load."""
        s = DatasetTransformation.schema()
        t = s.loads(
            """
            name: test
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.cols == "all"

    def test_schema_loads_false_output(self) -> None:
        """Tests schema load fails if output is False."""
        s = DatasetTransformation.schema()
        with pytest.raises(ValueError):
            s.loads(
                """
                name: test
                output: False
                """
            )

    def test_schema_load_fails_cols_nonstring(self) -> None:
        """Tests schema load fails if cols argument is not a string."""
        s = DatasetTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                cols: 1
                """
            )


def gen_name_test_dataset(cls: Type[Transformation], registry_name: str) -> None:
    """Tests name generation for dataset transformation classes.

    Test that a dataset transformation class is created with the correctly generated
    random name.
    """
    gen_name_test(cls, registry_name, cols="cols:A")


@unit_test
class TestCleanDataTransformation:
    """Tests for CleanDataTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_dataset(CleanDataTransformation, "cleandata")

    def test_registration(self) -> None:
        """Tests transformations registration."""
        registration_test(CleanDataTransformation, "cleandata")


@unit_test
class TestNormalizeDataTransformation:
    """Tests for NormalizeDataTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test_dataset(NormalizeDataTransformation, "normalizedata")

    def test_registration(self) -> None:
        """Tests transformations registration."""
        registration_test(NormalizeDataTransformation, "normalizedata")



"""Tests for utils.py."""
from unittest.mock import Mock

import pytest

from bitfount.schemas.exceptions import SchemaClassError
from bitfount.schemas.utils import (
    _combine_dict_helper,
    _dict_to_schema,
    _obj_to_schema,
    bf_dump,
)
from tests.utils.helper import unit_test


@unit_test
def test_dump_with_empty_object_error() -> None:
    """Tests that empty object dump raises error."""
    with pytest.raises(SchemaClassError):
        bf_dump(None)


@unit_test
def test_optional_nested_set() -> None:
    """Test that `optional_nested` is set on the class."""
    obj1 = Mock()
    obj1.optional_nested = None
    obj2 = Mock()
    obj2.fields_dict = {}
    obj2.nested_fields = {}
    obj2.optional_nested = ["blah"]
    _combine_dict_helper(obj1, obj2, {}, {})
    assert obj1.optional_nested == ["blah"]


@unit_test
def test_depth_greater_than5_raises_error_dict_to_schema() -> None:
    """Tests that depth greater than 5 raises error."""
    with pytest.raises(SchemaClassError):
        _dict_to_schema(Mock(), Mock(), depth=6)


@unit_test
def test_depth_greater_than5_raises_error_obj_to_schema() -> None:
    """Tests that depth greater than 5 raises error."""
    with pytest.raises(SchemaClassError):
        _obj_to_schema(Mock(), depth=6)



"""References to custom models."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Optional, Type, Union, cast

from marshmallow import fields

from bitfount.data.datastructure import DataStructure
from bitfount.data.datastructure import registry as datastructure_registry
from bitfount.data.schema import BitfountSchema
from bitfount.federated.logging import _get_federated_logger
from bitfount.hub.exceptions import ModelUploadError
from bitfount.hub.helper import _default_bitfounthub
from bitfount.models.base_models import _BaseModelRegistryMixIn
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    DistributedModelProtocol,
    _BaseSerializableObjectMixIn,
    _StrAnyDict,
)
from bitfount.utils import (
    _get_non_abstract_classes_from_module,
    _handle_fatal_error,
    delegates,
)

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub
    from bitfount.models.bitfount_model import BitfountModel

logger = _get_federated_logger(__name__)


@delegates()
class BitfountModelReference(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
    """Describes a local or remote reference to a `BitfountModel` class.

    :::tip

    To use another user's custom model, simply provide that user's username instead of
    your own (along with the name of the model as the `model_ref` argument).

    :::

    Args:
        model_ref: Either path to model file or name of model on hub.
        datastructure: `DataStructure` to be passed to the model when initialised.
        model_version: The version of the model you wish to use. Defaults to
            latest version.
        schema: The `BitfountSchema` object associated with the datasource
            on which the model will be trained on.
        username: The username of the model owner. Defaults to bitfount session username
            if not provided.
        hub: Required for upload/download of model. This attribute is set after
            initialisation on the worker side as the hub is not serialized. Defaults to
            None.
        hyperparameters: Hyperparameters to be passed to the model constructor after it
            has been loaded from file or hub. Defaults to None.

    Raises:
        ValueError: If `username` is not provided and `hub` is not provided.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "model_ref": fields.Method(
            serialize="get_model_ref", deserialize="load_model_ref"
        ),
        "model_version": fields.Int(allow_none=True),
        # The hub should not be serialized but can be deserialized if provided
        "hub": fields.Raw(allow_none=True, load_only=True),
        "username": fields.Str(allow_none=True),
        "hyperparameters": fields.Dict(keys=fields.Str()),
        "param_clipping": fields.Dict(
            keys=fields.String(), values=fields.Integer(), allow_none=True
        ),
        "schema": fields.Nested(BitfountSchema._Schema),
    }
    # TODO: [BIT-1954] Maybe this should actually fall under the hyperparameters
    #  rather than populating the top level with additional fields. Or at least
    #  a `other_kwargs` (terrible name, don't use that) so that we can add more
    #  to it in the future.
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"datastructure": datastructure_registry}

    def __init__(
        self,
        model_ref: Union[Path, str],
        datastructure: DataStructure,
        model_version: Optional[int] = None,
        schema: Optional[BitfountSchema] = None,
        username: Optional[str] = None,
        hub: Optional[BitfountHub] = None,
        hyperparameters: Optional[_StrAnyDict] = None,
    ):
        self.class_name = type(self).__name__
        self.model_ref = model_ref
        self.datastructure = datastructure
        self.model_version = model_version
        self.schema = schema if schema else BitfountSchema()
        self.hub = _default_bitfounthub(hub, username=username)
        self.hyperparameters = hyperparameters if hyperparameters is not None else {}
        self.username = username or self.hub.username

    def _get_model_from_path(self) -> Type[BitfountModel]:
        """Returns model class from path.

        Returns:
            The model class.
        """
        self.model_ref = cast(Path, self.model_ref)
        return _get_non_abstract_classes_from_module(self.model_ref)[
            self.model_ref.stem
        ]

    def _upload_model_to_hub(self) -> None:
        """Uploads model to hub under the logged-in user's account."""
        # model_ref is path to model code file
        self.model_ref = cast(Path, self.model_ref)
        try:
            self.hub.send_model(self.model_ref)
        except ModelUploadError as ex:
            _handle_fatal_error(ex)

    def _get_model_from_hub(self) -> Type[BitfountModel]:
        """Returns model class from hub from user denoted by `self.username`.

        Returns:
            The model class.
        """
        # model_ref is the name of a model on the hub
        self.model_ref = cast(str, self.model_ref)
        model_cls = self.hub.get_model(
            self.username, self.model_ref, self.model_version
        )

        # Check that the model has been retrieved correctly
        if not model_cls:
            raise ValueError(
                "Unable to retrieve model from hub, check logs for details."
            )
        return model_cls

    def get_model(self) -> Type[BitfountModel]:
        """Gets the model referenced.

        If the model is a Path to a `BitfountModel`, it will upload it to BitfountHub
        and return the model class. If it is a name of a model on the hub, it will
        download the model from the hub and return the model class.

        Returns:
            The model class.

        Raises:
            TypeError: If the model is not a Path or a string.
            TypeError: If the model does not implement `DistributedModelProtocol`.
            ValueError: If a `BitfountHub` instance has not been provided or if there
                was a communication error with the hub.
        """
        if isinstance(self.model_ref, Path):
            model_cls = self._get_model_from_path()

            # Check that chosen model is compatible with federation by checking if it
            # implements `DistributedModelProtocol`. The only way to do this is to
            # instantiate the model and perform an `isinstance` check.
            model = model_cls(
                datastructure=self.datastructure,
                schema=self.schema,
                **self.hyperparameters,
            )
            if not isinstance(model, DistributedModelProtocol):
                raise TypeError(
                    f"Model {self.model_ref.stem} does not implement "
                    f"DistributedModelProtocol."
                )

            self._upload_model_to_hub()

            # self.model_ref is set to the name of the model so that the model doesn't
            # get unnecessarily re-uploaded if `get_model` is called multiple times
            self.model_ref = self.model_ref.stem
        elif isinstance(self.model_ref, str):
            model_cls = self._get_model_from_hub()
        else:
            raise TypeError(f"Model of type {type(self.model_ref)} not recognised.")

        return model_cls



"""SQL query algorithm."""
from __future__ import annotations

from typing import Any, ClassVar, Dict, Mapping, Optional, cast

from marshmallow import fields
import pandas as pd
import pandasql

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.exceptions import DuplicateColumnError
from bitfount.federated.algorithms.base import (
    _BaseAlgorithmFactory,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.mixins import _ModellessAlgorithmMixIn
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.types import _DataLessAlgorithm
from bitfount.types import T_FIELDS_DICT
from bitfount.utils import delegates

logger = _get_federated_logger(__name__)


class _ModellerSide(_BaseModellerAlgorithm):
    """Modeller side of the SqlQuery algorithm."""

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Nothing to initialise here."""
        pass

    def run(self, results: Mapping[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
        """Simply returns results."""
        return dict(results)


class _WorkerSide(_BaseWorkerAlgorithm):
    """Worker side of the SqlQuery algorithm."""

    def __init__(
        self, *, query: str, table: Optional[str] = None, **kwargs: Any
    ) -> None:
        self.datasource: BaseSource
        self.query = query
        self.table = table
        super().__init__(**kwargs)

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_identifier: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Sets Datasource."""
        self.datasource = datasource
        if (
            self.table is None
            and pod_identifier is not None
            and not self.datasource.multi_table
        ):
            self.table = pod_identifier.split("/")[1]

    def run(self) -> pd.DataFrame:
        """Executes the query on the data source and returns a dataframe."""
        logger.info("Executing query...")
        if self.datasource.multi_table and isinstance(self.datasource, DatabaseSource):
            # Connect to the db directly if you are working with a multitable.
            conn = self.datasource.con.connect()
            output = pd.read_sql(sql=self.query, con=conn)
        elif self.datasource.multi_table and self.table is None:
            raise ValueError(
                "No table specified on which to execute the query on. "
                "Please specify the table on which to execute the query "
                "in the algorithm definition."
            )
        else:
            # For SQL queries on a dataframe/ single table.
            self.datasource.load_data(table_name=self.table)
            df = self.datasource.data
            if (f"from `{self.table}`" not in self.query) and (
                f"FROM `{self.table}`" not in self.query
            ):
                err_msg = """The default table for single table datasource is the pod
                    identifier without the username, in between backticks(``).
                    Please ensure your SQL query operates on that table. The
                    table name should be put inside backticks(``) in the
                    query statement, to make sure it is correctly parsed."""
                logger.error(err_msg)
                raise ValueError(err_msg)
            # Table name should be set by this point for single
            # table datasource, so it's safe to cast
            self.table = cast(str, self.table)
            # We need to remove hyphens as pandasql errors out if
            # they are included in the table name and query.
            query = self.query.split("`")
            query[query.index(self.table)] = query[query.index(self.table)].replace(
                "-", ""
            )
            table_name = self.table.replace("-", "")
            try:
                # Now for the actual query.
                output = pandasql.sqldf("".join(query), {table_name: df})
            except pandasql.PandaSQLException as ex:
                raise ValueError(
                    f"Error executing SQL query: [{self.query}], got error [{ex}]"
                )

        if any(output.columns.duplicated()):
            raise DuplicateColumnError(
                f"The following column names are duplicated in the output "
                f"dataframe: {output.columns[output.columns.duplicated()]}. "
                f"Please rename them in the query, and try again."
            )
        return cast(pd.DataFrame, output)


@delegates()
class SqlQuery(_BaseAlgorithmFactory, _ModellessAlgorithmMixIn, _DataLessAlgorithm):
    r"""Simple algorithm for running a SQL query on a table.

    :::info

    The default table for single-table datasources is the pod identifier without the
    username, in between backticks(\`\`). Please ensure your SQL query operates on
    that table. The table name should be put inside backticks(\`\`) in the query
    statement, to make sure it is correctly parsed e.g. ``SELECT MAX(G) AS MAX_OF_G
    FROM `df` ``. This is the standard quoting mechanism used by MySQL (and also
    included in SQLite).

    :::

    :::info

    If you are using a multi-table datasource, ensure that your SQL query syntax matches
    the syntax required by the Pod database backend.

    :::

    Args:
        query: The SQL query to execute.
        table: The target table name. For single table pod datasources,
            this will default to the pod name.

    Attributes:
        query: The SQL query to execute.
        table: The target table name. For single table pod datasources,
            this will default to the pod name.

    """

    def __init__(self, *, query: str, table: Optional[str] = None, **kwargs: Any):
        super().__init__()
        self.query = query
        self.table = table

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "query": fields.Str(),
        "table": fields.Str(allow_none=True),
    }

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the SqlQuery algorithm."""
        return _ModellerSide(**kwargs)

    def worker(self, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the SqlQuery algorithm."""
        return _WorkerSide(query=self.query, table=self.table, **kwargs)



"""Utility functions for the runner modules."""
from datetime import datetime
import logging
import os
from pathlib import Path
import sys
from typing import List, Optional

from bitfount.config import BITFOUNT_LOG_TO_FILE, BITFOUNT_LOGS_DIR

logger = logging.getLogger(__name__)


def setup_loggers(
    loggers: List[logging.Logger], name: Optional[str] = None
) -> List[logging.Logger]:
    """Set up loggers with console and file handlers.

    Creates a logfile in 'logs' directory with the current date and time and outputs all
    logs at the "DEBUG" level. Also outputs logs to stdout at the "INFO" level. A common
    scenario is to attach handlers only to the root logger, and to let propagation take
    care of the rest.

    Args:
        loggers: The logger(s) to setup
        name: Creates a subdirectory inside BITFOUNT_LOGS_DIR
            if provided. Defaults to None.

    Returns:
        A list of updated logger(s).
    """
    handlers: List[logging.Handler] = []

    # Check if logging to file is not disabled
    if BITFOUNT_LOG_TO_FILE:
        # Create directory if it doesn't exist
        parent_logfile_dir = (
            Path(os.getenv("BITFOUNT_LOGS_DIR", ".")) / BITFOUNT_LOGS_DIR
        )
        logfile_dir = parent_logfile_dir if not name else parent_logfile_dir / name
        logfile_dir.mkdir(parents=True, exist_ok=True)

        # Set file logging configuration
        file_handler = logging.FileHandler(
            f"{logfile_dir}/{datetime.now():%Y-%m-%d-%H%M%S}.log"
        )
        file_log_formatter = logging.Formatter(
            "%(asctime)s %(thread)-12d [%(levelname)-8s] %(name)s: %(message)s"
        )
        file_handler.setFormatter(file_log_formatter)
        file_handler.setLevel(logging.DEBUG)
        handlers.append(file_handler)

    # Set console logging configuration
    console_handler = logging.StreamHandler(sys.stdout)
    console_log_formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s]: %(message)s", datefmt="%H:%M:%S"
    )
    console_handler.setFormatter(console_log_formatter)
    console_handler.setLevel(logging.INFO)
    handlers.append(console_handler)

    # Cannot use `logger` as iter-variable as shadows outer name.
    for i_logger in loggers:
        # Clear any existing configuration
        list(map(i_logger.removeHandler, i_logger.handlers))
        list(map(i_logger.removeFilter, i_logger.filters))

        # Set base level to DEBUG and ensure messages are not duplicated
        i_logger.setLevel(logging.DEBUG)
        i_logger.propagate = False

        # Add handlers to loggers
        list(map(i_logger.addHandler, handlers))

    return loggers



"""Tests for the results only protocol."""
from pathlib import Path
import re
from typing import Callable, cast
from unittest.mock import Mock, create_autospec

from _pytest.logging import LogCaptureFixture
import pandas as pd
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount import (
    BitfountHub,
    BitfountSchema,
    ModelEvaluation,
    PrivateSqlQuery,
    SqlQuery,
)
from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.algorithms.column_avg import ColumnAverage
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.algorithms.model_algorithms.train_and_evaluate import (
    ModelTrainingAndEvaluation,
)
from bitfount.federated.protocols import results_only
import bitfount.federated.protocols.base as protocols
from bitfount.federated.protocols.base import (
    _BaseModellerProtocol,
    _BaseProtocol,
    _BaseWorkerProtocol,
)
from bitfount.federated.protocols.results_only import (
    ResultsOnly,
    _ModellerSide,
    _ResultsOnlyCompatibleAlgoFactory_,
    _ResultsOnlyCompatibleModelAlgoFactory,
    _ResultsOnlyCompatibleModellerAlgorithm,
    _ResultsOnlyDataCompatibleWorkerAlgorithm,
    _ResultsOnlyDataIncompatibleWorkerAlgorithm,
    _WorkerSide,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import (
    _InterPodWorkerMailbox,
    _WorkerMailbox,
)
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils import PytestRequest
from tests.utils.helper import (
    backend_test,
    create_datasource,
    create_datastructure,
    unit_test,
)


@unit_test
class TestWorkerSide:
    """Tests worker-side of ResultsOnly."""

    @fixture
    def mock_datasource(self) -> Mock:
        """Mock BaseSource."""
        mock_datasource: Mock = create_autospec(BaseSource, instance=True)
        return mock_datasource

    @fixture
    def mock_mailbox(self) -> Mock:
        """Mock WorkerMailbox."""
        mock_mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
        return mock_mailbox

    @fixture
    def worker_side_factory(self, mock_mailbox: Mock) -> Callable[[Mock], _WorkerSide]:
        """Factory to create WorkerSide instances from mock algorithms."""

        def _create(algo: Mock) -> _WorkerSide:
            return _WorkerSide(algorithm=algo, mailbox=mock_mailbox, aggregator=None)

        return _create

    async def test_run_with_data_algo(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests WorkerSide.run() with an algorithm needing data."""
        mock_data_algorithm: Mock = create_autospec(
            _ResultsOnlyDataCompatibleWorkerAlgorithm, instance=True
        )
        worker_side: _WorkerSide = worker_side_factory(mock_data_algorithm)

        await worker_side.run(mock_datasource)

        mock_data_algorithm.initialise.assert_called_once()
        mock_data_algorithm.run.assert_called_once_with(data=mock_datasource)
        mock_mailbox.send_evaluation_results.assert_awaited_once_with(
            mock_data_algorithm.run.return_value
        )
        mock_mailbox.get_task_complete_update.assert_called_once()

    async def test_run_with_non_data_algo(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests WorkerSide.run() with an algorithm not needing data."""
        mock_algorithm: Mock = create_autospec(
            _ResultsOnlyDataIncompatibleWorkerAlgorithm, instance=True
        )
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)

        await worker_side.run(mock_datasource)

        mock_algorithm.initialise.assert_called_once()
        mock_algorithm.run.assert_called_once_with()
        mock_mailbox.send_evaluation_results.assert_awaited_once_with(
            mock_algorithm.run.return_value
        )

    async def test_run_with_model_algorithm(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests WorkerrSide.run() receives parameters if model algo."""
        mock_algorithm: Mock = create_autospec(_BaseWorkerModelAlgorithm, instance=True)
        mock_algorithm.model = Mock()
        mock_algorithm.run = Mock()
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)
        mock_receive_parameters = mocker.patch.object(
            worker_side, "_receive_parameters"
        )

        await worker_side.run(mock_datasource)

        mock_receive_parameters.assert_awaited_once()
        mock_algorithm.initialise.assert_called_once()
        mock_algorithm.run.assert_called_once()
        mock_mailbox.send_evaluation_results.assert_awaited_once_with(
            mock_algorithm.run.return_value
        )

    async def test_receive_parameters(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Test _receive_parameters awaits _get_model_parameters."""
        mock_algorithm: Mock = create_autospec(
            _ResultsOnlyDataIncompatibleWorkerAlgorithm, instance=True
        )
        mock_get_model_parameters = mocker.patch(
            "bitfount.federated.protocols.results_only._get_model_parameters"
        )
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)

        await worker_side._receive_parameters()

        mock_get_model_parameters.assert_awaited_once_with(mock_mailbox)


@unit_test
class TestModellerSide:
    """Tests modeller-side of ResultsOnly."""

    @fixture
    def mock_datasource(self) -> Mock:
        """Mock DataSource."""
        mock_datasource: Mock = create_autospec(BaseSource, instance=True)
        return mock_datasource

    @fixture
    def mock_mailbox(self) -> Mock:
        """Mock ModellerMailbox."""
        mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
        return mock_mailbox

    @fixture
    def modeller_side_factory(
        self, mock_mailbox: Mock
    ) -> Callable[[Mock], _ModellerSide]:
        """Factory to create ModellerSide instances from mock algorithms."""

        def _create(algo: Mock) -> _ModellerSide:
            return _ModellerSide(algorithm=algo, mailbox=mock_mailbox, aggregator=None)

        return _create

    async def test_run(
        self,
        mock_mailbox: Mock,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide.run() with an algorithm not needing data."""
        mock_algorithm: Mock = create_autospec(
            _ResultsOnlyCompatibleModellerAlgorithm, instance=True
        )
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)

        await modeller_side.run()

        mock_algorithm.initialise.assert_called_once()
        mock_algorithm.run.assert_called_once()
        mock_mailbox.get_evaluation_results_from_workers.assert_awaited_once()

    async def test_run_with_model_algorithm(
        self,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide.run() sends parameters if model algo."""
        mock_algorithm: Mock = create_autospec(
            _BaseModellerModelAlgorithm, instance=True
        )
        mock_algorithm.model = Mock()
        mock_algorithm.run = Mock()
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)
        mock_send_parameters = mocker.patch.object(modeller_side, "_send_parameters")

        await modeller_side.run()

        mock_send_parameters.assert_awaited_once()
        mock_algorithm.initialise.assert_called_once()
        mock_algorithm.run.assert_called_once()
        mock_mailbox.get_evaluation_results_from_workers.assert_awaited_once()

    async def test_send_parameters(
        self,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Test _send_parameters awaits _send_model_parameters."""
        mock_algorithm: Mock = create_autospec(
            _ResultsOnlyDataIncompatibleWorkerAlgorithm, instance=True
        )
        new_parameters = Mock()
        mock_send_model_parameters = mocker.patch(
            "bitfount.federated.protocols.results_only._send_model_parameters"
        )
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)

        await modeller_side._send_parameters(new_parameters)

        mock_send_model_parameters.assert_awaited_once_with(
            new_parameters, mock_mailbox
        )


@unit_test
class TestResultsOnly:
    """Test Results Only protocol."""

    @fixture(scope="function", params=[None, "/mock/file"])
    def remote_algorithm(self, request: PytestRequest) -> ModelTrainingAndEvaluation:
        """Returns remote algorithm."""
        return ModelTrainingAndEvaluation(model=Mock(), pretrained_file=request.param)

    def test_init_method_with_aggregator(
        self, remote_algorithm: ModelTrainingAndEvaluation
    ) -> None:
        """Tests init method with aggregator."""
        aggregator: Mock = Mock()
        protocol = ResultsOnly(algorithm=remote_algorithm, aggregator=aggregator)
        assert protocol.aggregator is aggregator

    def test_init_method_with_secure_aggregation(
        self, remote_algorithm: ModelTrainingAndEvaluation
    ) -> None:
        """Tests init method with aggregator."""
        protocol = ResultsOnly(algorithm=remote_algorithm, secure_aggregation=True)
        assert isinstance(protocol.aggregator, SecureAggregator)

    def test_init_method_without_aggregator(
        self, caplog: LogCaptureFixture, remote_algorithm: ModelTrainingAndEvaluation
    ) -> None:
        """Tests init method with aggregator."""
        caplog.set_level("INFO")
        protocol = ResultsOnly(algorithm=remote_algorithm)
        assert protocol.aggregator is None
        assert (
            "No aggregator specified. Will return a dictionary of results."
            in caplog.text
        )

    def test_modeller(
        self, mock_modeller_mailbox: Mock, remote_algorithm: ModelTrainingAndEvaluation
    ) -> None:
        """Test modeller method."""
        protocol_factory = ResultsOnly(algorithm=remote_algorithm)
        protocol = protocol_factory.modeller(mailbox=mock_modeller_mailbox)
        for type_ in [
            _BaseProtocol,
            _BaseModellerProtocol,
            results_only._ModellerSide,
        ]:
            assert isinstance(protocol, type_)

    def test_worker(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        remote_algorithm: ModelTrainingAndEvaluation,
    ) -> None:
        """Test worker method."""
        protocol_factory = ResultsOnly(algorithm=remote_algorithm)
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
        for type_ in [
            _BaseProtocol,
            _BaseWorkerProtocol,
            results_only._WorkerSide,
        ]:
            assert isinstance(protocol, type_)

    def test_worker_with_aggregator(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        remote_algorithm: ModelTrainingAndEvaluation,
    ) -> None:
        """Test worker method with aggregator."""
        aggregator = Mock(spec=Aggregator)
        protocol_factory = ResultsOnly(
            algorithm=remote_algorithm, aggregator=aggregator
        )
        protocol: _WorkerSide = protocol_factory.worker(
            mailbox=mock_worker_mailbox, hub=mock_hub
        )
        assert protocol.aggregator is not None
        aggregator.worker.assert_called_once()

    async def test_worker_with_aggregator_privatesql_results(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Test worker method with aggregator."""
        aggregator_factory = Aggregator()

        algorithm_factory = PrivateSqlQuery(
            query="SELECT AVG(G) AS AVG_OF_G FROM df.df",
            epsilon=0.1,
            delta=0.001,
            column_ranges={},
        )
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        protocol_factory = ResultsOnly(
            algorithm=algorithm_factory, aggregator=aggregator_factory
        )
        df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        results = {
            "pod1": (
                {
                    "used_epsilon": 0.1,
                    "used_delta": 0.001,
                    "column_epsilon": 0.1,
                    "column_delta": 0.001,
                },
                df,
            )
        }
        mocker.patch.object(algorithm, "run", return_value=results)
        protocol: _WorkerSide = protocol_factory.worker(
            mailbox=mock_worker_mailbox, hub=mock_hub
        )
        mock_aggregator_run = mocker.patch.object(protocol.aggregator, "run")
        await protocol.run(create_datasource(classification=True))
        assert protocol.aggregator is not None
        mock_aggregator_run.assert_called_once()

    def test_worker_with_secure_aggregator(
        self,
        mock_hub: Mock,
        remote_algorithm: ModelTrainingAndEvaluation,
    ) -> None:
        """Test worker method with secure aggregator."""
        aggregator = Mock(spec=SecureAggregator)
        mailbox = Mock(spec=_InterPodWorkerMailbox)
        protocol_factory = ResultsOnly(
            algorithm=remote_algorithm, aggregator=aggregator
        )

        protocol: _WorkerSide = protocol_factory.worker(mailbox=mailbox, hub=mock_hub)
        assert protocol.aggregator is not None
        aggregator.worker.assert_called_once_with(mailbox=mailbox)

    def test_worker_with_secure_aggregator_raises_type_error(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        remote_algorithm: ModelTrainingAndEvaluation,
    ) -> None:
        """Test worker method with secure aggregator raises TypeError.

        This happens if the mailbox is not compatible with the aggregator.
        """
        aggregator = Mock(spec=SecureAggregator)
        protocol_factory = ResultsOnly(
            algorithm=remote_algorithm, aggregator=aggregator
        )
        with pytest.raises(
            TypeError,
            match="Inter-pod aggregators require an inter-pod worker mailbox.",
        ):
            protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)

    def test_worker_with_unrecognised_aggregator_raises_type_error(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        remote_algorithm: ModelTrainingAndEvaluation,
    ) -> None:
        """Test worker method with unrecognised aggregator raises TypeError."""
        aggregator = Mock()
        protocol_factory = ResultsOnly(
            algorithm=remote_algorithm, aggregator=aggregator
        )
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"Unrecognised aggregator factory ({type(aggregator)}); "
                f"unable to determine how to call .worker() factory method."
            ),
        ):
            protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)

    def test__validate_algorithm_accepts(self) -> None:
        """Tests _validate_algorithm accepts compatible."""
        # Test with ResultsOnlyCompatibleAlgoFactory
        mock_algorithm: Mock = create_autospec(
            _ResultsOnlyCompatibleAlgoFactory_, instance=True
        )
        ResultsOnly._validate_algorithm(mock_algorithm)

        # Test with ResultsOnlyCompatibleModelAlgoFactory
        mock_algorithm = create_autospec(
            _ResultsOnlyCompatibleModelAlgoFactory, instance=True
        )
        ResultsOnly._validate_algorithm(mock_algorithm)

    def test__validate_algorithm_rejects(self) -> None:
        """Tests _validate_algorithm rejects incompatible."""
        mock_algorithm: Mock = Mock(spec_set=["__name__"])
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"The {ResultsOnly.__name__} protocol does not "
                f"support the {type(mock_algorithm).__name__} algorithm."
            ),
        ):
            ResultsOnly._validate_algorithm(mock_algorithm)

    @unit_test
    def test_run_protocol(self, mocker: MockerFixture) -> None:
        """Tests we can run a protocol."""
        algorithm = ColumnAverage(field="TARGET", table_name="fake")
        # Mock out Modeller creation
        mock_modeller = mocker.patch(
            "bitfount.federated.protocols.base._Modeller", autospec=True
        )
        mock_modeller.return_value = mock_modeller  # for __init__
        mock_modeller.run.return_value = None

        protocol = ResultsOnly(algorithm=algorithm)
        protocol.run(
            pod_identifiers=["fake/fake"],
            hub=Mock(),
            message_service=Mock(),
            private_key_or_file=Path("fake.pem"),
        )

        mock_modeller.run.assert_called_once_with(["fake/fake"], require_all_pods=False)

    @unit_test
    def test_run_protocol_idp_is_set(self, mocker: MockerFixture) -> None:
        """Tests that the idp url gets initialized."""
        algorithm = ColumnAverage(field="TARGET", table_name="fake")

        # Mock out Modeller creation
        mock_modeller = mocker.patch(
            "bitfount.federated.protocols.base._Modeller", autospec=True
        )
        mock_modeller.return_value = mock_modeller  # for __init__
        mock_modeller.run.return_value = None
        mock_idp_url = mocker.patch(
            "bitfount.federated.protocols.base._get_idp_url",
            return_value="https://idp-url.unit-test.bitfount.com",
        )

        protocol = ResultsOnly(algorithm=algorithm)
        protocol.run(
            pod_identifiers=["fake/fake"],
            hub=Mock(),
            message_service=Mock(),
            private_key_or_file=Path("fake.pem"),
        )

        mock_idp_url.assert_called_once()


@backend_test
@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for column average algorithm."""

    def test_serialization_model_algorithm(self) -> None:
        """Test Marshmallow Serialization for a model algorithm."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=2,
        )
        algorithm_factory = ModelEvaluation(model=model)
        results_only = ResultsOnly(algorithm=algorithm_factory)
        dumped = bf_dump(results_only)
        loaded = bf_load(dumped, protocols.registry)
        assert results_only.class_name == loaded.class_name
        from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal

        assert_vars_equal(
            vars(
                cast(
                    _ResultsOnlyCompatibleModelAlgoFactory, results_only.algorithm
                ).model
            ),
            vars(loaded.algorithm.model),
        )

    def test_serialization_sql_algorithm(self) -> None:
        """Test Marshmallow Serialization for sql query algorithm."""
        algorithm_factory = SqlQuery(query="SELECT * FROM df")
        results_only = ResultsOnly(algorithm=algorithm_factory)
        dumped = bf_dump(results_only)
        loaded = bf_load(dumped, protocols.registry)
        assert results_only.class_name == loaded.class_name
        assert results_only.algorithm.__dict__ == loaded.algorithm.__dict__



"""Tests federated secure module."""
from bitfount import SecureShare
from bitfount.federated.secure import (
    FLOAT_32_BIT_PRECISION,
    LARGE_PRIME_NUMBER,
    _secure_share_registry,
)
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import unit_test


def asert_vars_equal(agg: object, dumped: object) -> None:
    """Helper function for comparing aggregators."""
    for item in vars(agg).keys():
        if item != "_tensor_shim":
            assert vars(agg)[item] == vars(dumped)[item]


@unit_test
class TestMarshmallowSerialization:
    """Tests serialization of SecureShare."""

    def test_dump_load(self) -> None:
        """Tests serialization of SecureShare."""
        aggregator_factory = SecureShare(
            prime_q=LARGE_PRIME_NUMBER,
            precision=FLOAT_32_BIT_PRECISION,
        )
        serialized_agg = bf_dump(aggregator_factory)
        deserialized_agg = bf_load(serialized_agg, _secure_share_registry)
        asert_vars_equal(deserialized_agg, deserialized_agg)



"""Tests `bitfount_model.py`."""
from pathlib import Path
from typing import List, Type

from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.models.bitfount_model import BitfountModel
from tests.utils.helper import unit_test


@unit_test
class TestBitfountModel:
    """Tests BitfountModel class."""

    @fixture
    def bitfount_model_class(self) -> Type[BitfountModel]:
        """Returns a BitfountModel instance."""

        class MyModel(BitfountModel):
            """A test model."""

            @staticmethod
            def _get_import_statements() -> List[str]:
                """Returns a list of import statements."""
                return ["import random_library"]

        return MyModel

    def test_serialize_model_source_code(
        self,
        bitfount_model_class: Type[BitfountModel],
        mocker: MockerFixture,
        tmp_path: Path,
    ) -> None:
        """Tests serialize_model_source_code method."""
        mock_os = mocker.patch("bitfount.models.bitfount_model.os")
        mock_isort = mocker.patch("bitfount.models.bitfount_model.isort")

        filename = tmp_path / "model.py"
        bitfount_model_class.serialize_model_source_code(
            filename, extra_imports=["from blah import blah"]
        )

        mock_os.system.assert_called()  # Called twice
        mock_isort.file.assert_called_once()

        with open(filename, "r") as f:
            model_code = f.read()
            assert "from blah import blah" in model_code
            assert "import random_library" in model_code



"""Typed wrappers around the Bitfount REST apis."""
from __future__ import annotations

from dataclasses import dataclass
import hashlib
import logging
from pathlib import Path
from typing import Any, Final, List, Mapping, Optional, Tuple, Type, cast
import warnings

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
import requests
from requests import HTTPError, RequestException, Response
from requests.exceptions import InvalidJSONError

from bitfount import web_utils
from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _PRODUCTION_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    _get_environment,
)
from bitfount.data.schema import BitfountSchema
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.types import SerializedProtocol, _PodResponseType
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.hub.exceptions import (
    ModelTooLargeError,
    ModelUploadError,
    ModelValidationError,
    SchemaUploadError,
)
from bitfount.hub.types import (
    _DEV_AM_URL,
    _STAGING_AM_URL,
    PRODUCTION_AM_URL,
    PRODUCTION_HUB_URL,
    _AccessManagerKeyResponseJSON,
    _AMAccessCheckResponseJSON,
    _ModelDetailsResponseJSON,
    _ModelUploadResponseJSON,
    _MonitorPostJSON,
    _MultiModelDetailsResponseJSON,
    _MultiPodDetailsResponseJSON,
    _OIDCAccessCheckPostJSON,
    _PodDetailsResponseJSON,
    _PodRegistrationResponseJSON,
    _SAMLAdditionalInfoPOSTJSON,
    _SAMLChallengeResponseJSON,
    _SignatureBasedAccessCheckPostJSON,
)
from bitfount.models.bitfount_model import BitfountModel
from bitfount.storage import (
    _download_data_from_s3,
    _download_file_from_s3,
    _get_packed_data_object_size,
    _upload_data_to_s3,
    _upload_file_to_s3,
)
from bitfount.types import (
    _JSONDict,
    _S3PresignedPOSTFields,
    _S3PresignedPOSTURL,
    _S3PresignedURL,
    _SAMLResponse,
)
from bitfount.utils import _get_mb_from_bytes, _get_non_abstract_classes_from_module

logger = logging.getLogger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950


# This should match the corresponding value in
# bitfount/bitfount-web:bf-packages/bitfount-hub/pages/api/pods.api.js
_MAX_SCHEMA_SIZE_BYTES: Final[int] = 10 * 1024 * 1024  # 10 Megabytes
_MAX_SCHEMA_SIZE_MEGABYTES: Final[int] = _get_mb_from_bytes(
    _MAX_SCHEMA_SIZE_BYTES
).whole
# This should match the corresponding value in
# bitfount/bitfount-web:bf-packages/bitfount-hub/pages/api/models.api.js
_MAX_CUSTOM_MODEL_SIZE_BYTES: Final[int] = 3 * 1024 * 1024  # 3 Megabytes
_MAX_CUSTOM_MODEL_SIZE_MEGABYTES: Final[int] = _get_mb_from_bytes(
    _MAX_CUSTOM_MODEL_SIZE_BYTES
).whole


@dataclass
class PodPublicMetadata:
    """Data about the Pod which is publicly visible on BitfountHub."""

    name: str
    display_name: str
    description: str
    schema: _JSONDict  # This should be set to schema.to_json()


def hash_file_contents(file_path: Path) -> str:
    """Hash file contents using sha256."""
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


def _verify_signature(
    public_key: RSAPublicKey, signature: bytes, message: bytes
) -> bool:
    """Verifies that decrypting `signature` with `public_key` == `message`."""
    try:
        public_key.verify(
            signature,
            message,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=20,  # padding.PSS.MAX_LENGTH,
            ),
            hashes.SHA256(),
        )
    except InvalidSignature:
        logger.warning("Signature invalid.")
        return False

    logger.info("Signature valid.")
    return True


def _handle_request_exception(
    exc: RequestException, response: Optional[Response], conn_type: str
) -> str:
    """Logs out the request exception and response details, if provided."""
    err_msg = f"Bitfount {conn_type} connection failed with: {exc}."
    # Explicit None check needed due to truthyness of Response objects
    if response is not None and response.text:
        err_msg += f" Response was: {response.text}"
    logger.error(err_msg)

    return err_msg


def _extract_response_json(response: Response) -> Any:
    """Extracts the JSON from the response, raising an HTTPError if there is none."""
    try:
        response_json = response.json()
    except ValueError:  # raised if JSON extraction fails
        raise InvalidJSONError(
            f"Invalid JSON response ({response.status_code}): {response.text}"
        )
    return response_json


def _check_pod_id_details(
    pod_identifier: Optional[str] = None,
    pod_namespace: Optional[str] = None,
    pod_name: Optional[str] = None,
) -> str:
    """Checks that either pod_identifier OR pod_namespace and pod_name are provided.

    If both sets are provided, then makes sure that the constructed pod_identifier
    matches the supplied one.

    Args:
        pod_identifier: Full pod identifier (i.e. <pod_namespace>/<pod_name>).
        pod_namespace: Pod namespace name.
        pod_name: Pod name.

    Returns:
        The pod_identifier.

    Raises:
        ValueError:
            - If no args are provided.
            - If pod_namespace is provided but pod_name is not (or vice versa).
            - If all args are provided but the constructed pod_identifier doesn't
              match the provided one.
    """
    args = [pod_identifier, pod_namespace, pod_name]

    # At least one set of args must be provided.
    if not any(args):
        raise ValueError(
            "At least one of pod_identifier OR "
            "pod_namespace and pod_name must be provided"
        )

    # Either both pod_namespace and pod_name are provided or both aren't
    if bool(pod_namespace) != bool(pod_name):
        raise ValueError(
            "Both pod_namespace and pod_name must be provided, or neither must be."
        )

    # If all are provided, the constructed and provided pod_identifiers must match
    if all(args):
        if pod_identifier != f"{pod_namespace}/{pod_name}":
            raise ValueError(
                "pod_identifier, pod_namespace and pod_name all provided, "
                "but pod_identifier doesn't match pod_namespace/pod_name."
            )

    # If conditions are met, return as needed
    if pod_identifier:
        return pod_identifier
    else:
        warnings.warn(
            "pod_identifier should be used instead of pod_namespace and pod_name",
            DeprecationWarning,
            stacklevel=3,
        )
        return f"{pod_namespace}/{pod_name}"


class BitfountHub:
    """A typed API for interacting with BitfountHub.

    Args:
        session: Bitfount session for authentication.
        url: URL to Bitfount Hub. Defaults to PRODUCTION_HUB_URL.
    """

    def __init__(
        self,
        session: Optional[BitfountSession] = None,
        url: str = PRODUCTION_HUB_URL,
    ):
        self.session = session if session else BitfountSession()
        if not self.session.authenticated:
            self.session.authenticate()
        self.url = url.rstrip("/")  # to ensure no trailing slash
        self.user_storage_path = self.session.user_storage_path
        self.username = self.session.username

    @staticmethod
    def _handle_hub_request_exception(
        err: RequestException, response: Optional[Response]
    ) -> str:
        """Logs out the request exception and response details, if provided."""
        return _handle_request_exception(err, response, "Hub")

    def get_pod(
        self,
        pod_identifier: Optional[str] = None,
        pod_namespace: Optional[str] = None,
        pod_name: Optional[str] = None,
    ) -> Optional[_PodDetailsResponseJSON]:
        """Gets the details of a pod.

        Either pod_identifier or pod_namespace and pod_name can be provided.

        Args:
            pod_identifier: Full pod identifier (i.e. <pod_namespace>/<pod_name>).
            pod_namespace: Pod namespace name.
            pod_name: Pod name.

        Returns:
            The pod details JSON or None if unable to retrieve them.
        """
        # We reuse the arg name here because we're guaranteeing it's set now
        pod_identifier: str = _check_pod_id_details(  # type: ignore[no-redef] # Reason: see comment # noqa: B950
            pod_identifier, pod_namespace, pod_name
        )
        # Delete the others to avoid incorrect usage
        del pod_namespace
        del pod_name

        # Retrieve pod details from hub
        try:
            logger.debug(f"Requesting pod information for '{pod_identifier}' from hub.")
            response = self.session.get(f"{self.url}/api/pods/{pod_identifier}")
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            self._handle_hub_request_exception(err, err_response)
            return None

        # Extract the JSON from the response body
        response_json: _PodDetailsResponseJSON = _extract_response_json(response)
        return response_json

    def get_all_pods(self) -> List[_MultiPodDetailsResponseJSON]:
        """Returns a list of all registered and visible pods.

        Returns:
            A list of JSON details for each visible pod or a empty list if unable
            to retrieve them.
        """
        try:
            response = self.session.get(f"{self.url}/api/pods")
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            self._handle_hub_request_exception(err, err_response)
            return []

        response_json: List[_MultiPodDetailsResponseJSON] = _extract_response_json(
            response
        )
        return response_json

    def get_all_models(self) -> List[_MultiModelDetailsResponseJSON]:
        """Returns a list of all registered models.

        Returns:
           A list of JSON details for each model or a empty list if unable
           to retrieve them.
        """
        try:
            response = self.session.get(f"{self.url}/api/models")
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            self._handle_hub_request_exception(err, err_response)
            return []

        response_json: List[_MultiModelDetailsResponseJSON] = _extract_response_json(
            response
        )
        return response_json

    def get_pod_key(
        self,
        pod_identifier: Optional[str] = None,
        pod_namespace: Optional[str] = None,
        pod_name: Optional[str] = None,
    ) -> Optional[str]:
        """Gets the public key of a pod.

        Either pod_identifier or pod_namespace and pod_name can be provided.

        Args:
            pod_identifier: Full pod identifier (i.e. <pod_namespace>/<pod_name>).
            pod_namespace: Pod namespace name.
            pod_name: Pod name.

        Returns:
            The pod key or None if no key can be found.

        Raises:
            HTTPError: If the response from the hub is malformed.
        """
        # Get pod details and check validity
        logger.debug(f"Requesting public key for '{pod_identifier}' from hub")
        pod_details: Optional[_PodDetailsResponseJSON] = self.get_pod(
            pod_identifier, pod_namespace, pod_name
        )

        if not pod_details:
            # Return None so that the caller is aware that
            # they aren't guaranteed a string key!
            return None

        # Extract key from response
        pod_key: Optional[str] = pod_details.get("podPublicKey")

        if not pod_key:
            logger.error(f"Response JSON contained no public key: {pod_details}")
            return None

        return pod_key

    def send_model(self, model_code_path: Path) -> None:
        """Sends the provided `model_code` to Hub, associated to session username.

        The name of the model is taken from the name of the file.

        Args:
            model_code_path: The path to the file containing the model code.

        Raises:
            ModelUploadError: If the model upload fails. Specific subclasses of
                ModelUploadError are used to indicate different failure conditions.
        """
        # Check model is correctly formed
        try:
            self._verify_bitfount_model_format(model_code_path)
        except (ValueError, TypeError) as e:
            logger.error(f"Model incorrectly structured. Error: {e}")
            raise ModelValidationError("Model incorrectly structured.") from e
        except ImportError as ie:
            logger.error(ie)
            raise ModelValidationError("Unable to import model.") from ie

        model_name: str = model_code_path.stem
        model_code_text: bytes = model_code_path.read_text().encode("utf-8")
        model_size = len(model_code_text)
        model_hash = hash_file_contents(model_code_path)
        if model_size > _MAX_CUSTOM_MODEL_SIZE_BYTES:
            raise ModelTooLargeError(
                f"Model is too large to upload: "
                f"expected max {_MAX_CUSTOM_MODEL_SIZE_MEGABYTES} megabytes, "
                f"got {_get_mb_from_bytes(model_size).fractional} megabytes."
            )

        # Register model details with hub
        try:
            logger.info(f"Uploading model '{model_name}' to Hub")
            response = self.session.post(
                f"{self.url}/api/models",
                json={
                    "modelName": model_name,
                    "modelSize": model_size,
                    "modelHash": model_hash,
                },
            )
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            self._handle_hub_request_exception(err, err_response)
            raise ModelUploadError(
                "Request exception occurred when uploading model details to hub."
            ) from err

        # Extract response
        response_json: _ModelUploadResponseJSON = _extract_response_json(response)
        if not response_json["success"]:
            logger.error(
                f"Could not send model to Bitfount Hub. "
                f"Failed with message: {response_json['errorMessage']}"
            )
            raise ModelUploadError(
                f"Failed to upload model details to hub: "
                f"{response_json['errorMessage']}"
            )
        logger.info(
            "Model successfully registered to Bitfount Hub, "
            "uploading to object storage."
        )

        # Extract URL to upload model to and upload it
        upload_url: _S3PresignedPOSTURL = response_json["uploadUrl"]
        upload_fields: _S3PresignedPOSTFields = response_json["uploadFields"]
        try:
            _upload_file_to_s3(
                upload_url,
                upload_fields,
                file_contents=model_code_text,
                file_name=model_code_path.name,
            )
        except RequestException as re:
            logger.error(f"Failed to upload model to S3: {re}")
            raise ModelUploadError("Failed to upload model to S3") from re

        logger.info("Model successfully uploaded to object storage.")
        return None

    @staticmethod
    def _verify_bitfount_model_format(path_to_model: Path) -> Type[BitfountModel]:
        """Verifies that the model file is correctly formatted.

        Args:
            path_to_model: Path to the model file.

        Returns:
            The parsed model class.

        Raises:
            ImportError: If the file could not be loaded.
            TypeError: If there is no BitfountModel subclass in the file, or it
                is still abstract.
            ValueError: If multiple model classes are contained in the file.
            ValueError: If the name of the class and the name of the file differ.

        """
        models = _get_non_abstract_classes_from_module(path_to_model)
        models = {
            name: class_
            for name, class_ in models.items()
            if issubclass(class_, BitfountModel)
        }

        num_models = len(models)
        if num_models == 0:
            raise TypeError(
                "Subclass of `BitfountModel` not found in file or is still abstract."
            )
        elif num_models > 1:
            raise ValueError(
                f"Model file contains {num_models} models. Must be just 1."
            )

        model = models.popitem()
        if model[0] != path_to_model.stem:
            raise ValueError(
                f"{model[0]} != {path_to_model.stem}. "
                "Model class name must be the same as the filename",
            )

        logger.debug("Model is formatted correctly")

        return model[1]

    def get_model(
        self, username: str, model_name: str, model_version: Optional[int] = None
    ) -> Optional[Type[BitfountModel]]:
        """Gets model code corresponding to `model_name` from user `username`.

        Args:
            username: The model's owner.
            model_name: The name of the model.
            model_version: (Optional) The version of the model. Defaults to latest.

        Returns:
            The loaded Bitfount model class or None if unable to retrieve it.
        """
        # Retrieve model details from API endpoint
        try:
            params = {
                "modellerName": username,
                "modelName": model_name,
            }
            if model_version:
                params.update({"modelVersion": str(model_version)})
            response = self.session.get(
                f"{self.url}/api/models",
                params=params,
            )
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            self._handle_hub_request_exception(err, err_response)
            return None

        # Check response contains model details
        response_json: _ModelDetailsResponseJSON = _extract_response_json(response)
        if len(response_json) == 0:  # i.e. empty JSON response
            logger.warning(
                f"No models registered by the name of {model_name} from user {username}"
            )
            return None

        # Download model code from S3
        try:
            model_url: _S3PresignedURL = response_json["modelDownloadUrl"]
        except (KeyError, TypeError):
            # Either entry isn't in the dict, or it's not a dict (might be None)
            raise InvalidJSONError(
                f"Cannot retrieve model, no model URL in pod response: "
                f"{response_json}"
            )

        # Extract class from downloaded code. We save it out to file to make it
        # easier to import and verify.
        # This will be a str because we provide an encoding arg.
        model_code: str = cast(str, _download_file_from_s3(model_url, encoding="utf-8"))

        model_file = self._write_model_code_to_file(
            model_code, self.user_storage_path, model_name
        )

        file_hash = hash_file_contents(model_file)
        if file_hash != response_json["modelHash"]:
            raise ModelValidationError(
                "Stored hash does not match hashed model file contents. "
                "Model must be re-uploaded to hub."
            )
        try:
            model_class = self._verify_bitfount_model_format(model_file)
        except (ValueError, TypeError, ImportError) as e:
            logger.error(e)
            # Remove the model file as model is incompatible
            model_file.unlink()
            return None
        return model_class

    @staticmethod
    def _write_model_code_to_file(
        model_code: str, user_storage_path: Path, model_name: str
    ) -> Path:
        """Saves downloaded model code to a Python file.

        Creates a module from `model_code` and saves it in a directory named after
        `username` under user_storage_path/models
        """
        model_store = user_storage_path / "models"
        model_store.mkdir(parents=True, exist_ok=True)
        model_file = Path(model_store / model_name).with_suffix(".py")
        model_file.touch()
        model_file.write_text(model_code)

        return model_file

    def register_pod(
        self,
        public_metadata: PodPublicMetadata,
        pod_public_key: RSAPublicKey,
        access_manager_key: RSAPublicKey,
    ) -> None:
        """Registers a pod with the hub.

        The pod's schema will be uploaded to S3.

        Args:
            public_metadata: Details about the pod (name, etc) to register.
            pod_public_key: The public key to use for this pod.
            access_manager_key: The public key of the Access Manager.

        Raises:
            RequestException: If unable to connect to the hub.
            HTTPError: If the response is not successful.
        """
        # Need to determine how large it will be once packed and uploaded to S3
        schema_size = _get_packed_data_object_size(public_metadata.schema)
        if schema_size > _MAX_SCHEMA_SIZE_BYTES:
            raise SchemaUploadError(
                f"Schema is too large to upload: "
                f"expected max {_MAX_SCHEMA_SIZE_MEGABYTES} megabytes, "
                f"got {_get_mb_from_bytes(schema_size).fractional} megabytes."
            )

        json_post = {
            "name": public_metadata.name,
            "podDisplayName": public_metadata.display_name,
            "podPublicKey": pod_public_key.public_bytes(
                Encoding.OpenSSH, PublicFormat.OpenSSH
            ).decode("utf-8"),
            "accessManagerPublicKey": access_manager_key.public_bytes(
                Encoding.OpenSSH, PublicFormat.OpenSSH
            ).decode("utf-8"),
            "description": public_metadata.description,
            "schemaSize": schema_size,
        }

        # Send pod registration request and get response from hub
        try:
            response = self.session.post(
                f"{self.url}/api/pods",
                json=json_post,
                timeout=10,
            )
            # We raise_for_status() here but also need to explicitly check that
            # it's a 200/201 status code later.
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            err_msg = self._handle_hub_request_exception(err, err_response)
            # Raise the same type of exception that was caught
            raise type(err)(err_msg) from err

        # Extract the JSON from the response body
        response_json: _PodRegistrationResponseJSON = _extract_response_json(response)

        # Handle non-successful status codes
        if response.status_code in (200, 201):
            logger.info("Pod registration successful")
        else:
            raise HTTPError(
                f"Unexpected response ({response.status_code}): {response.text}"
            )

        # Extract URL to upload schema to and upload it
        logger.info("Uploading schema to S3")
        upload_url: _S3PresignedPOSTURL = response_json["uploadUrl"]
        upload_fields: _S3PresignedPOSTFields = response_json["uploadFields"]
        _upload_data_to_s3(upload_url, upload_fields, public_metadata.schema)

    def do_pod_heartbeat(self, pod_name: str, pod_public_key: RSAPublicKey) -> None:
        """Makes a "heartbeat" update to the hub.

        This takes the form of a PATCH request to the `/api/pods` endpoint.

        Args:
            pod_name: The name of the pod.
            pod_public_key: The public key of the pod.
        """
        json_patch = {
            "name": pod_name,
            "podPublicKey": pod_public_key.public_bytes(
                Encoding.OpenSSH, PublicFormat.OpenSSH
            ).decode("utf-8"),
        }

        # We don't need to catch RequestExceptions here as the calling loop
        # handles that.
        response = self.session.patch(
            f"{self.url}/api/pods",
            json=json_patch,
            timeout=10,
        )
        # We do raise_for_status() here but also need to explicitly check
        # if it is status_code 200 below.
        response.raise_for_status()

        # Check that we received valid JSON (though we don't use it)
        try:
            _extract_response_json(response)
        except InvalidJSONError as err:
            logger.warning(err)
            return

        if response.status_code == 200:
            logger.debug("Status ping successful")
        else:
            logger.warning(
                f"Status update failed with ({response.status_code}): {response.text}"
            )

    def get_pod_schema(self, pod_identifier: str) -> BitfountSchema:
        """Retrieves the schema for a given pod.

        Args:
            pod_identifier: The pod to retrieve the schema of.

        Returns:
            The BitfountSchema for that pod.

        Raises:
            HTTPError: If there is an issue in the response.
        """
        logger.info(f'Retrieving schema for "{pod_identifier}"')
        pod_details: Optional[_PodDetailsResponseJSON] = self.get_pod(pod_identifier)

        # Try and extract schema URL from response
        try:
            if pod_details:
                schema_url: _S3PresignedURL = pod_details["schemaDownloadUrl"]
            else:
                # i.e. pod_details is None
                raise TypeError("No pod details retrieved from Hub")
        except (KeyError, TypeError):
            # Either entry isn't in the dict, or it's not a dict (might be None)
            raise InvalidJSONError(
                f"Cannot retrieve pod schema, no schema URL in pod response: "
                f"{pod_details}"
            )

        # Download and parse schema
        schema_json: _JSONDict = _download_data_from_s3(schema_url)
        return BitfountSchema.load(schema_json)

    def send_monitor_update(self, update: _MonitorPostJSON) -> None:
        """Send an update to the monitor service.

        Args:
            update: The monitor service update to send as a JSON dict.

        Raises:
            RequestException: If unable to connect to the hub.
            HTTPError: If the response is not successful.
        """
        try:
            response = self.session.post(
                f"{self.url}/api/ingest", json=update, timeout=2
            )
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            err_msg = self._handle_hub_request_exception(err, err_response)
            # Raise the same type of exception that was caught
            raise type(err)(err_msg) from err


class BitfountAM:
    """Typed API for communicating with the Bitfount Access Manager.

    Args:
        session: Bitfount session for authentication.
        access_manager_url: URL of the access manager.
    """

    def __init__(
        self,
        session: Optional[BitfountSession] = None,
        access_manager_url: Optional[str] = None,
    ):
        # Set the AM url based on the environment
        access_manager_url = (
            access_manager_url if access_manager_url else self._get_am_url()
        )
        self.session = session if session else BitfountSession()
        if not self.session.authenticated:
            self.session.authenticate()
        self.access_manager_url = access_manager_url.rstrip(
            "/"
        )  # to ensure no trailing slashes

    @staticmethod
    def _handle_am_request_exception(
        err: RequestException, response: Optional[Response]
    ) -> str:
        """Logs out the request exception and response details, if provided."""
        return _handle_request_exception(err, response, "Access Manager")

    @staticmethod
    def _get_am_url() -> str:
        """Gets the AM url based on the environment."""
        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            access_manager_url = _STAGING_AM_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            access_manager_url = _DEV_AM_URL
        elif environment == _PRODUCTION_ENVIRONMENT:
            access_manager_url = PRODUCTION_AM_URL
        return access_manager_url

    def get_access_manager_key(self) -> RSAPublicKey:
        """Gets the Access Manager's public key.

        Returns:
            The access manager's public key.

        Raises:
            RequestException: If there is a problem communicating with the
                              access manager.
        """
        try:
            # No need to use `self.session` here because the endpoint does not require
            # authentication.
            logger.debug(
                f"Retrieving access manager public key from "
                f"{self.access_manager_url}/api/access-manager-key"
            )
            response = web_utils.get(
                f"{self.access_manager_url}/api/access-manager-key",
                timeout=5,
            )
            logger.debug(f"Access manager public key response received: {response}")
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            err_msg = self._handle_am_request_exception(err, err_response)
            # Raise same type of exception as caught before
            raise type(err)(err_msg) from err

        response_json: _AccessManagerKeyResponseJSON = _extract_response_json(response)

        try:
            am_key: str = response_json["accessManagerPublicKey"]
            logger.debug(f"Received access manager key: {am_key}")
        except (KeyError, TypeError):
            raise InvalidJSONError(
                f"Unable to extract public key from access manager response, "
                f"no key in JSON: {response_json}"
            )
        return _RSAEncryption.load_public_key(am_key.encode())

    def get_saml_challenge(self) -> Tuple[str, str]:
        """Gets a fresh SAML challenge for a user.

        Returns:
            A tuple of:
                - The SAML Request (Challenge).
                - A request ID that the Pod should keep.

        Raises:
            RequestException: If there is a problem communicating with the
                              access manager.
            HTTPError: If it is unable to extract the SAML challenge from
                       the response.
        """
        try:
            response = self.session.get(
                f"{self.access_manager_url}/api/saml?idp=cli", timeout=10
            )
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            err_msg = self._handle_am_request_exception(err, err_response)
            # Raise same type of exception as caught before
            raise type(err)(err_msg) from err

        response_json: _SAMLChallengeResponseJSON = _extract_response_json(response)

        try:
            saml_request_to_proxy_to_modeller: str = response_json["samlRequest"]
            saml_request_id: str = response_json["id"]
        except (KeyError, TypeError):
            raise InvalidJSONError(
                f"Unable to extract SAML Challenge from access manager response, "
                f"no challenge in JSON: {response_json}"
            )
        return saml_request_to_proxy_to_modeller, saml_request_id

    def validate_saml_response(
        self,
        saml_response: _SAMLResponse,
        saml_request_id: str,
        pod_identifier: str,
        modeller_name: str,
        serialized_protocol: SerializedProtocol,
    ) -> _PodResponseType:
        """Check if modeller task should be accepted.

        This sends the SAML Response from a modeller,
        along with details about their training request
        to the access manager so that it can inform the pod
        whether or not to accept the job.

        Args:
            saml_response: SAML Response from Modeller
            saml_request_id: The ID of the original SAML request. The pod should have
                stored this.
            pod_identifier: The pod identifier of the pod that the task is for
            modeller_name: The username of the modeller that has sent the task
            serialized_protocol: The serialized protocol to use for the task received
                from the modeller.

        Returns:
            Response code from the Access manager.

        Raises:
            RequestException: If there is a problem communicating with the
                access manager.
            HTTPError: If it is unable to extract the JSON from the response.
            InvalidJSONError: If JSON does not match expected format.
        """
        additional_info = _SAMLAdditionalInfoPOSTJSON(
            originalSAMLRequestID=saml_request_id,
            podIdentifier=pod_identifier,
            modellerName=modeller_name,
            modellerProtocolRequest=serialized_protocol,
            identityProvider="SAML",
        )
        post_json: _JSONDict = dict(saml_response, **additional_info)

        return self._do_authorization_check(
            f"{self.access_manager_url}/api/access?idp=cli", post_json
        )

    def check_oidc_access_request(
        self,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
        modeller_name: str,
        modeller_access_token: str,
    ) -> _PodResponseType:
        """Check access using access token from OIDC request.

        Args:
            pod_identifier: The pod identifier for the request.
            serialized_protocol: The protocol received from the modeller.
            modeller_name: The name of the modeller.
            modeller_access_token: OAuth access token for the modeller.

        Returns:
            Response code from the Access manager.

        Raises:
            RequestException: If there is a problem communicating with the
                access manager.
            HTTPError: If it is unable to extract the JSON from the response.
            InvalidJSONError: If JSON does not match expected format.
        """
        post_json = _OIDCAccessCheckPostJSON(
            podIdentifier=pod_identifier,
            modellerProtocolRequest=serialized_protocol,
            modellerName=modeller_name,
            modellerToken=modeller_access_token,
            identityProvider="OIDC",
        )

        return self._do_authorization_check(
            f"{self.access_manager_url}/api/access", post_json
        )

    def check_signature_based_access_request(
        self,
        unsigned_task: bytes,
        signed_task: bytes,
        pod_identifier: str,
        serialized_protocol: SerializedProtocol,
        modeller_name: str,
    ) -> _PodResponseType:
        """Check access by verifying the signed task against the unsigned task.

        Check that when modeller_name signs the unsigned task it matches the signed
        version of the task.

        Args:
            unsigned_task: Unsigned task.
            signed_task: Signed task.
            pod_identifier: The pod identifier for the request.
            serialized_protocol: The protocol received from the modeller.
            modeller_name: The name of the modeller.
            modeller_access_token: OAuth access token for the modeller.

        Returns:
            Response code from the Access manager.

        Raises:
            RequestException: If there is a problem communicating with the
                access manager.
            HTTPError: If it is unable to extract the JSON from the response.
            InvalidJSONError: If JSON does not match expected format.
        """
        post_json = _SignatureBasedAccessCheckPostJSON(
            podIdentifier=pod_identifier,
            modellerRequest=serialized_protocol,
            modellerName=modeller_name,
            unsignedTask=unsigned_task,
            signedTask=signed_task,
            identityProvider="SIGNATURE",
        )

        return self._do_authorization_check(
            f"{self.access_manager_url}/api/access", post_json
        )

    def _do_authorization_check(
        self,
        url: str,
        post_json: Mapping[str, Any],
    ) -> _PodResponseType:
        """Makes an authorization check to the Access Manager.

        Args:
            url: AM URL to POST to.
            post_json: The JSON to send.

        Returns:
            Response code from the Access manager.

        Raises:
            RequestException: If there is a problem communicating with the
                access manager.
            HTTPError: If it is unable to extract the JSON from the response.
            InvalidJSONError: If JSON does not match expected format.
        """
        try:
            response = self.session.post(
                url=url,
                timeout=10,
                json=post_json,
            )
            response.raise_for_status()
        except RequestException as err:
            try:
                # noinspection PyUnboundLocalVariable
                err_response: Optional[Response] = response
            except NameError:
                # response variable wasn't created yet
                err_response = None
            err_msg = self._handle_am_request_exception(err, err_response)
            # Raise same type of exception as caught before
            raise type(err)(err_msg) from err

        try:
            response_json: _AMAccessCheckResponseJSON = _extract_response_json(response)
            logger.debug(f"JSON From access manager: {response_json}")
            try:
                return _PodResponseType[response_json["code"]]
            except KeyError:
                return _PodResponseType.NO_ACCESS
            except TypeError:
                raise AttributeError()
        except AttributeError:
            # raised if response is not a dict
            raise InvalidJSONError(
                f"Invalid JSON response ({response.status_code}): {response.text}"
            )



"""PyTorch-specific DataLoader implementations."""
import math
import random
import secrets
from typing import Any, Iterator, List, Union, cast

import torch
from torch.utils.data import DataLoader as PyTorchDataLoader

from bitfount.backends.pytorch.data.datasets import (
    _PyTorchDataset,
    _PyTorchIterableDataset,
)
from bitfount.backends.pytorch.data.utils import _convert_batch_to_tensor
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.types import _DataBatch, _SingleOrMulti
from bitfount.utils import delegates

#: The default buffer size for shuffling Iterable Datasets.
DEFAULT_BUFFER_SIZE: int = 1000


class _BasePyTorchBitfountDataLoader(BitfountDataLoader):
    """Base class for PyTorch-specific Bitfount DataLoaders.

    Args:
       dataset: An pytorch compatible dataset.
       batch_size: The batch size for the dataloader.
           Defaults to 1.
       shuffle: A boolean value indicating whether the values
           in the dataset should be shuffled. Defaults to False.

    Attributes:
       dataset: An pytorch compatible dataset.
       batch_size: The batch size for the dataloader.
           Defaults to 1.
       shuffle: A boolean value indicating whether the values
           in the dataset should be shuffled. Defaults to False.
    """

    def __init__(
        self,
        dataset: Union[_PyTorchDataset, _PyTorchIterableDataset],
        batch_size: int = 1,
        shuffle: bool = False,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle


@delegates()
class PyTorchBitfountDataLoader(_BasePyTorchBitfountDataLoader):
    """Wraps a PyTorch DataLoader with bitfount functions.

    Args:
       dataset: An pytorch compatible dataset.
    """

    def __init__(self, dataset: _PyTorchDataset, **kwargs: Any):
        super().__init__(dataset=dataset, **kwargs)
        self.dataloader = self.get_pytorch_dataloader()

    def get_pytorch_dataloader(self, **kwargs: Any) -> PyTorchDataLoader:
        """Return a PyTorch DataLoader for `self.dataset`.

        Keyword arguments are passed to PyTorch's DataLoader constructor and take
        precedence over the values set in the constructor.
        """
        return PyTorchDataLoader(
            dataset=kwargs.pop("dataset", self.dataset),
            batch_size=kwargs.pop("batch_size", self.batch_size),
            shuffle=kwargs.pop("shuffle", self.shuffle),
            **kwargs,
        )

    def __len__(self) -> int:
        """Number of batches or number of elements if batch size is None."""
        return len(self.dataloader)

    def __iter__(self) -> Iterator[List[_SingleOrMulti[torch.Tensor]]]:
        """Wrapper around the default PyTorch DataLoader iterator.

        The only difference is that the elements of each batch are wrapped in a list.
        """
        for batch in self.dataloader:
            yield [x for x in batch]


@delegates()
class PyTorchIterableBitfountDataLoader(_BasePyTorchBitfountDataLoader):
    """Wraps a PyTorch DataLoader with bitfount functions.

    Args:
        dataset: An iterable dataset.
        secure_rng: A boolean value indicating whether to use a secure
            random number generator. Defaults to False.

    Attributes:
         secure_rng: A boolean value indicating whether to use a secure
            random number generator. Defaults to False.
    """

    def __init__(
        self, dataset: _PyTorchIterableDataset, secure_rng: bool = False, **kwargs: Any
    ):
        super().__init__(dataset=dataset, **kwargs)
        self.secure_rng = secure_rng

    @property
    def buffer_size(self) -> int:
        """Number of elements to buffer.

        The size of the buffer is the greater of the batch size and default buffer size
        unless the dataset is smaller than the default buffer in which case the dataset
        size is used. PyTorch already ensures that the batch size is not greater than
        the dataset size under the hood.
        """
        # Batch size is optional in the core hierarchy but in pytorch we ensure it is
        # set to 1 if not provided. Re-assuring mypy of this.
        assert self.batch_size is not None  # nosec[assert_used]
        return max(min(len(self.dataset), DEFAULT_BUFFER_SIZE), self.batch_size)

    def __len__(self) -> int:
        """Number of batches in the dataset."""
        assert self.batch_size is not None  # nosec[assert_used]
        return math.ceil(len(self.dataset) / self.batch_size)

    def __iter__(self) -> Iterator[List[_SingleOrMulti[torch.Tensor]]]:
        """Yields a batch of data when iterated.

        We use a custom iterator with different behaviour depending on whether the
        dataset should be shuffled or not. Each batch is explicitly converted to torch
        tensors prior to yielding as this is not done automatically by pytorch.
        """
        batch: _DataBatch = []

        if self.shuffle:
            # If the dataset should be shuffled, we use a reservoir sampling method
            # to sample from a buffer of elements from the dataset.
            buffer: _DataBatch = []
            for sample in cast(_PyTorchIterableDataset, self.dataset):
                if len(batch) == self.batch_size:
                    yield _convert_batch_to_tensor(batch)
                    batch = []

                if len(buffer) == self.buffer_size:
                    if self.secure_rng:
                        idx = secrets.randbelow(self.buffer_size)
                    else:
                        # Ignoring security warning here because RNG does not need
                        # to be cryptographically secure if it is turned off by
                        # the user.
                        idx = random.randint(0, self.buffer_size - 1)  # nosec
                    batch.append(buffer[idx])
                    buffer[idx] = sample
                else:
                    buffer.append(sample)

            # This is only reached once the dataset iterator has been exhausted. The
            # remainder of the buffer is shuffled and yielded until empty.
            random.shuffle(buffer)
            while buffer:
                if len(batch) == self.batch_size:
                    yield _convert_batch_to_tensor(batch)
                    batch = []

                batch.append(buffer.pop())

        else:
            # If the dataset should not be shuffled, we simply iterate over the dataset
            for sample in cast(_PyTorchIterableDataset, self.dataset):
                if len(batch) == self.batch_size:
                    yield _convert_batch_to_tensor(batch)
                    batch = []

                batch.append(sample)

        # If there are any elements left in the batch after the dataset/buffer are
        # empty, yield an incomplete batch.
        if batch:
            yield _convert_batch_to_tensor(batch)



"""Decorator versions of functions in monitor.py."""
from functools import wraps
from typing import Any, Callable, TypeVar, cast

from bitfount.federated.monitoring.monitor import task_status_update
from bitfount.federated.monitoring.types import MonitorRecordPrivacy

_F = TypeVar("_F", bound=Callable[..., Any])


def task_status(
    message: str, privacy: MonitorRecordPrivacy = MonitorRecordPrivacy.ALL_PARTICIPANTS
) -> Callable[[_F], _F]:
    """A decorator that will send a TASK_STATUS_UPDATE message at function start."""

    def _wrapper(func: _F) -> _F:
        @wraps(func)
        def _wrapped(*args: Any, **kwargs: Any) -> Any:
            task_status_update(message=message, privacy=privacy)
            return func(*args, **kwargs)

        return cast(_F, _wrapped)

    return _wrapper
